r"""
The :mod:`pyunlocbox.acceleration` module implements acceleration schemes for
use with the :mod:`pyunlocbox.solvers`. Pass a given acceleration object as an
argument to your chosen solver during its initialization so that the solver can
use it.

Interface
---------

The :class:`accel` base class defines a common interface to all acceleration
schemes:

.. autosummary::

    accel.pre
    accel.update_step
    accel.update_sol
    accel.post

Acceleration schemes
--------------------

Then, derived classes implement various common acceleration schemes.

.. autosummary::

    dummy
    backtracking
    fista
    fista_backtracking
    regularized_nonlinear

"""

import copy
import logging
import warnings

import numpy as np

# Note: line_search_armijo moved to private module in scipy >= 1.9
# TODO: Consider migrating to scipy.optimize.line_search for future use
try:
    from scipy.optimize.linesearch import line_search_armijo
except ImportError:
    from scipy.optimize._linesearch import line_search_armijo


class accel:
    r"""
    Defines the acceleration scheme object interface.

    This class defines the interface of an acceleration scheme object intended
    to be passed to a solver inheriting from
    :class:`pyunlocbox.solvers.solver`. It is intended to be a base class for
    standard acceleration schemes which will implement the required methods.
    It can also be instantiated by user code and dynamically modified for
    rapid testing. This class also defines the generic attributes of all
    acceleration scheme objects.

    """

    def __init__(self):
        pass

    def pre(self, functions, x0):
        """
        Pre-processing specific to the acceleration scheme.

        Gets called when :func:`pyunlocbox.solvers.solve` starts running.
        """
        self._pre(functions, x0)

    def _pre(self, functions, x0):
        raise NotImplementedError("Class user should define this method.")

    def update_step(self, solver, objective, niter):
        """
        Update the step size for the next iteration.

        Parameters
        ----------
        solver : pyunlocbox.solvers.solver
            Solver on which to act.
        objective : list of floats
            List of evaluations of the objective function since the beginning
            of the iterative process.
        niter : int
            Current iteration number.

        Returns
        -------
        float
            Updated step size.

        """
        return self._update_step(solver, objective, niter)

    def _update_step(self, solver, objective, niter):
        raise NotImplementedError("Class user should define this method.")

    def update_sol(self, solver, objective, niter):
        """
        Update the solution point for the next iteration.

        Parameters
        ----------
        solver : pyunlocbox.solvers.solver
            Solver on which to act.
        objective : list of floats
            List of evaluations of the objective function since the beginning
            of the iterative process.
        niter : int
            Current iteration number.

        Returns
        -------
        array_like
            Updated solution point.

        """
        return self._update_sol(solver, objective, niter)

    def _update_sol(self, solver, objective, niter):
        raise NotImplementedError("Class user should define this method.")

    def post(self):
        """
        Post-processing specific to the acceleration scheme.

        Mainly used to delete references added during initialization so that
        the garbage collector can free the memory. Gets called when
        :func:`pyunlocbox.solvers.solve` finishes running.

        """
        self._post()

    def _post(self):
        raise NotImplementedError("Class user should define this method.")


class dummy(accel):
    r"""
    Dummy acceleration scheme which does nothing.

    Used by default in most of the solvers. It simply returns unaltered the
    step size and solution point it receives.

    """

    def _pre(self, functions, x0):
        pass

    def _update_step(self, solver, objective, niter):
        return solver.step

    def _update_sol(self, solver, objective, niter):
        return solver.sol

    def _post(self):
        pass


# -----------------------------------------------------------------------------
# Stepsize optimizers
# -----------------------------------------------------------------------------


class backtracking(dummy):
    r"""
    Backtracking acceleration based on a local quadratic approximation of the
    smooth part of the objective.

    Parameters
    ----------
    eta : float
        A number between 0 and 1 representing the ratio of the geometric
        sequence formed by successive step sizes. In other words, it
        establishes the relation `step_new = eta * step_old`.
        Default is 0.5.

    Notes
    -----
    This is the backtracking strategy used in the original FISTA paper,
    :cite:`beck2009FISTA`.

    Examples
    --------
    >>> import numpy as np
    >>> from pyunlocbox import functions, solvers, acceleration
    >>> y = [4, 5, 6, 7]
    >>> x0 = np.zeros(len(y))
    >>> f1 = functions.norm_l1(y=y, lambda_=1.0)
    >>> f2 = functions.norm_l2(y=y, lambda_=0.8)
    >>> accel = acceleration.backtracking()
    >>> solver = solvers.forward_backward(accel=accel, step=10)
    >>> ret = solvers.solve([f1, f2], x0, solver, atol=1e-32, rtol=None)
    ... # doctest: +ELLIPSIS
    Solution found after ... iterations:
        objective function f(sol) = 0.000000e+00
        stopping criterion: ATOL
    >>> ret['sol']
    array([4., 5., 6., 7.])

    """

    def __init__(self, eta=0.5, **kwargs):
        if (eta > 1) or (eta <= 0):
            raise ValueError("eta must be between 0 and 1.")
        self.eta = eta
        super().__init__(**kwargs)

    def _update_step(self, solver, objective, niter):
        """
        Notes
        -----
        TODO: For now we're recomputing gradients in order to evaluate the
        backtracking criterion. In the future, it might be interesting to
        think of some design changes so that this function has access to the
        gradients directly.

        Since the call to `solver._algo()` modifies the internal state of the
        solver itself, we need to store the solver's property values before
        doing backtracking, and then restore the solver's state after
        backtracking is done. This takes more memory, but it's the only way to
        guarantee that backtracking is performing on a fixed solver state.
        """
        # Save current state of the solver
        properties = copy.deepcopy(vars(solver))
        logging.debug(f"(Begin) solver properties: {properties}")

        # Initialize some useful variables
        fn = 0
        grad = np.zeros_like(properties["sol"])
        for f in solver.smooth_funs:
            fn += f.eval(properties["sol"])
            grad += f.grad(properties["sol"])
        step = properties["step"]

        logging.debug(f"fn = {fn}")

        while True:
            # Run the solver with the current stepsize
            solver.step = step
            logging.debug(f"Current step: {step}")
            solver._algo()
            logging.debug(f"(During) solver properties: {vars(solver)}")

            # Record results
            fp = np.sum([f.eval(solver.sol) for f in solver.smooth_funs])
            logging.debug(f"fp = {fp}")

            dot_prod = np.dot(solver.sol - properties["sol"], grad)
            logging.debug(f"dot_prod = {dot_prod}")

            norm_diff = np.sum((solver.sol - properties["sol"]) ** 2)
            logging.debug(f"norm_diff = {norm_diff}")

            # Restore the previous state of the solver
            for key, val in properties.items():
                setattr(solver, key, copy.copy(val))
            logging.debug(f"(Reset) solver properties: {vars(solver)}")

            if 2.0 * step * (fp - fn - dot_prod) <= norm_diff:
                logging.debug("Break condition reached")
                break
            else:
                logging.debug("Decreasing step")
                step *= self.eta

        return step


# -----------------------------------------------------------------------------
# Solution point optimizers
# -----------------------------------------------------------------------------


class fista(dummy):
    r"""
    FISTA acceleration for forward-backward solvers.

    Notes
    -----
    This is the acceleration scheme proposed in the original FISTA paper,
    :cite:`beck2009FISTA`.

    Examples
    --------
    >>> import numpy as np
    >>> from pyunlocbox import functions, solvers, acceleration
    >>> y = [4, 5, 6, 7]
    >>> x0 = np.zeros(len(y))
    >>> f1 = functions.norm_l2(y=y)
    >>> f2 = functions.dummy()
    >>> accel = acceleration.fista()
    >>> solver = solvers.forward_backward(accel=accel, step=0.5)
    >>> ret = solvers.solve([f1, f2], x0, solver, atol=1e-5)
    Solution found after 15 iterations:
        objective function f(sol) = 4.957288e-07
        stopping criterion: ATOL
    >>> ret['sol']
    array([4.0002509 , 5.00031362, 6.00037635, 7.00043907])

    """

    def __init__(self, **kwargs):
        self.t = 1.0
        super().__init__(**kwargs)

    def _pre(self, functions, x0):
        self.sol = np.array(x0, copy=True)

    def _update_sol(self, solver, objective, niter):
        self.t = 1.0 if (niter == 1) else self.t  # Restart variable t if needed
        t = (1.0 + np.sqrt(1.0 + 4.0 * self.t**2.0)) / 2.0
        y = solver.sol + ((self.t - 1) / t) * (solver.sol - self.sol)
        self.t = t
        self.sol[:] = solver.sol
        return y

    def _post(self):
        del self.sol


class regularized_nonlinear(dummy):
    r"""
    Regularized nonlinear acceleration (RNA) for gradient descent solvers.

    Parameters
    ----------
    k : int, optional
        Number of points to keep in the buffer for computing the extrapolation.
        (Default is 10.)
    lambda_ : float or list of floats, optional
        Regularization parameter in the acceleration scheme. The user can pass
        a list of candidates, and the acceleration algorithm will pick the one
        that provides the best extrapolation.
        (Default is 1e-6.)
    adaptive : boolean, optional
        If adaptive = True and the user has not provided a list of
        regularization parameters, the acceleration algorithm will assemble a
        grid of possible regularization parameters based on the SVD of the
        Gram matrix of vectors of differences in the extrapolation buffer.
        If adaptive = False, the algorithm will simply try to use the value(s)
        given in `lambda_`.
        (Default is True.)
    dolinesearch : boolean, optional
        If dolinesearch = True, the acceleration scheme will try to return a
        point in the line segment between the current extrapolation and the
        previous one that provides a decrease in the value of the objective
        function.
        If dolinesearch = False, the algorithm simply returns the current
        extrapolation.
        (Default is True.)
    forcedecrease : boolean, optional
        If forcedecrese = True and we obtain a bad extrapolation, the
        algorithm returns the unchanged solution produced by the solver.
        If forcedecrease = False, the algorithm returns the new extrapolation
        no matter what.
        (Default is True.)

    Notes
    -----
    This is the acceleration scheme proposed in :cite:`scieur2016`.

    See also Damien Scieur's `repository <https://github.com/windows7lover
    /RegularizedNonlinearAcceleration>`_ for the Matlab version that inspired
    this implementation.

    Examples
    --------
    >>> import numpy as np
    >>> from pyunlocbox import functions, solvers, acceleration
    >>> dim = 25
    >>> np.random.seed(0)
    >>> xstar = np.random.rand(dim)  # True solution.
    >>> x0 = np.random.rand(dim)
    >>> x0 = xstar + 5*(x0 - xstar) / np.linalg.norm(x0 - xstar)
    >>> A = np.random.rand(dim, dim)
    >>> step = 1 / np.linalg.norm(np.dot(A.T, A))
    >>> f = functions.norm_l2(lambda_=0.5, A=A, y=np.dot(A, xstar))
    >>> fd = functions.dummy()
    >>> accel = acceleration.regularized_nonlinear(k=5)
    >>> solver = solvers.gradient_descent(step=step, accel=accel)
    >>> params = {'rtol':0, 'maxit':200, 'verbosity':'NONE'}
    >>> ret = solvers.solve([f, fd], x0, solver, **params)
    >>> pctdiff = 100 * np.sum((xstar - ret['sol'])**2) / np.sum(xstar**2)
    >>> print('Difference: {0:.1f}%'.format(pctdiff))
    Difference: 1.3%

    """

    def __init__(
        self,
        k=10,
        lambda_=1e-6,
        adaptive=True,
        dolinesearch=True,
        forcedecrease=True,
        **kwargs,
    ):
        self.k = k
        self.lambda_ = lambda_
        self.adaptive = adaptive
        self.dolinesearch = dolinesearch
        self.forcedecrease = forcedecrease

    @property
    def lambda_(self):
        return self._lambda_

    @lambda_.setter
    def lambda_(self, lambda_):
        try:
            self._lambda_ = [float(elem) for elem in lambda_]
        except TypeError:
            try:
                self._lambda_ = [float(lambda_)]
            except ValueError as err:
                print("lambda_ is not a number")
                raise err
        except ValueError as err:
            print("lambda_ is not a list of numbers")
            raise err

    def _pre(self, functions, x0):
        self.buffer = []
        self.functions = functions

    def _update_sol(self, solver, objective, niter):

        if (niter % (self.k + 1)) == 0:  # Extrapolate at each k iterations

            self.buffer.append(solver.sol)

            # (Normalized) matrix of differences
            U = np.diff(self.buffer, axis=0)
            UU = np.dot(U, U.T)
            UU /= np.linalg.norm(UU)

            # If no parameter grid was provided, assemble one.
            if self.adaptive and (len(self.lambda_) <= 1):
                svals = np.sort(np.abs(np.linalg.eigvals(UU)))
                svals = np.log(svals)
                svals = 0.5 * (svals[:-1] + svals[1:])
                self.lambda_ = np.concatenate(([0.0], np.exp(svals)))

            # Grid search for the best parameter for the extrapolation
            fvals = []
            c = np.zeros((self.k,))
            extrap = np.zeros_like(solver.sol)

            for lambda_ in self.lambda_:
                # Coefficients of the extrapolation
                c[:] = np.linalg.solve(UU + lambda_ * np.eye(self.k), np.ones(self.k))
                c[:] /= np.sum(c)

                extrap[:] = np.dot(np.asarray(self.buffer[:-1]).T, c)

                fvals.append(np.sum([f.eval(extrap) for f in self.functions]))

            if self.forcedecrease and (min(fvals) > np.sum(objective[-1])):
                # If we have bad extrapolations, keep solution as is
                extrap[:] = solver.sol
            else:
                # Return the best extrapolation from the grid search
                lambda_ = self.lambda_[fvals.index(min(fvals))]

                # We can afford to solve the linear system here again because
                # self.k is normally very small. Alternatively, we could have
                # kept track of the best extrapolations during the grid search,
                # but that would require at least double the memory, as we'd
                # have to store both the current extrapolation and the best
                # extrapolation.
                c[:] = np.linalg.solve(UU + lambda_ * np.eye(self.k), np.ones(self.k))
                c[:] /= np.sum(c)
                extrap[:] = np.dot(np.asarray(self.buffer[:-1]).T, c)

            # Improve proposal with line search
            if self.dolinesearch:
                # Objective evaluation functional
                def f(x):
                    return np.sum([f.eval(x) for f in self.functions])

                # Solution at previous extrapolation
                xk = self.buffer[0]
                # Search direction
                pk = extrap - xk
                # Objective value during the previous extrapolation
                old_fval = np.sum(objective[-self.k])

                a, fc, fa = line_search_armijo(
                    f=f, xk=xk, pk=pk, gfk=-pk, old_fval=old_fval, c1=1e-4, alpha0=1.0
                )

                # New point proposal
                if a is None:
                    warnings.warn("Line search failed to find good step size")
                else:
                    extrap[:] = xk + a * pk

            # Clear buffer and parameter grid for next extrapolation process
            self.buffer = []
            self.lambda_ = [] if self.adaptive else self.lambda_

            return extrap

        else:  # Gather points for future extrapolation
            self.buffer.append(copy.copy(solver.sol))
            return solver.sol

    def _post(self):
        del self.buffer, self.functions


# -----------------------------------------------------------------------------
# Mixed optimizers
# -----------------------------------------------------------------------------


class fista_backtracking(backtracking, fista):
    r"""
    FISTA acceleration with backtracking for forward-backward solvers.

    Notes
    -----
    This is the acceleration scheme and backtracking strategy proposed in the
    original FISTA paper, :cite:`beck2009FISTA`.

    Examples
    --------
    >>> import numpy as np
    >>> from pyunlocbox import functions, solvers, acceleration
    >>> y = [4, 5, 6, 7]
    >>> x0 = np.zeros(len(y))
    >>> f1 = functions.norm_l2(y=y)
    >>> f2 = functions.dummy()
    >>> accel = acceleration.fista_backtracking()
    >>> solver = solvers.forward_backward(accel=accel, step=0.5)
    >>> ret = solvers.solve([f1, f2], x0, solver, atol=1e-5)
    Solution found after 15 iterations:
        objective function f(sol) = 4.957288e-07
        stopping criterion: ATOL
    >>> ret['sol']
    array([4.0002509 , 5.00031362, 6.00037635, 7.00043907])

    """

    def __init__(self, eta=0.5, **kwargs):
        # I can do multiple inheritance here and avoid the deadly diamond of
        # death because the classes backtracking and fista modify different
        # methods of their parent class dummy. If that would not be the case, I
        # guess the best solution would be to inherit from accel and rewrite
        # the `_update_step()` and `_update_sol()` methods.

        backtracking.__init__(self, eta=eta, **kwargs)
        fista.__init__(self, **kwargs)



r"""
SV_DECONV
=====
This example implements the saptially varying PSF deconvolution algorithm
based on: Flicker & Rigaut, 2005 https://doi.org/10.1364/JOSAA.22.000504
"""

import numpy as np
from scipy.signal import fftconvolve

import pyunlocbox as plx

###############################################################################
# Sample data
im = np.random.randn(2748, 3840)
W = np.random.randn(20, np.prod(im.shape))
U = np.random.randn(np.prod(im.shape), 20)


# Setup Forward and Adjoint Models and Create Solver
def forward(im):
    im_sim = np.zeros_like(im)
    for i in range(15):
        weight = W[i, :].reshape(*im.shape)
        psf_mode = U[:, i].reshape(*im.shape)
        im_sim += fftconvolve(im * weight, psf_mode, mode="same")
    return im_sim


def forward_adj(im):
    im_sim = np.zeros_like(im)
    for i in range(15):
        weight = W[i, :].reshape(*im.shape)
        psf_mode = U[:, i].reshape(*im.shape)
        im_sim += fftconvolve(im, np.flipud(np.fliplr(psf_mode)), mode="same") * weight
    return im_sim


tau = 10
f1 = plx.functions.norm_l2(y=im, A=forward, At=forward_adj, lambda_=tau)
f2 = plx.functions.norm_tv(maxit=50, dim=2)
solver = plx.solvers.forward_backward(step=0.5 / tau, accel=plx.acceleration.fista())
ret = plx.solvers.solve(
    [f1, f2], x0=im.copy(), solver=solver, maxit=10, verbosity="ALL"
)



r"""
The package is mainly organized around two class hierarchies: the functions and
the solvers. Instantiated functions represent convex functions to optimize.
Instantiated solvers represent solving algorithms. The
:func:`pyunlocbox.solvers.solve` solving function takes as parameters a solver
object and some function objects to actually solve the optimization problem.
See this function's documentation for a typical usage example.

The :mod:`pyunlocbox` package is divided into the following modules:

* :mod:`.functions`: objective functions to define an optimization problem,
* :mod:`.solvers`: the main solving function and common solvers,
* :mod:`.acceleration`: general acceleration schemes for various solvers,
* :mod:`.operators`: some operators.

"""

from pyunlocbox import acceleration  # noqa: F401
from pyunlocbox import functions  # noqa: F401
from pyunlocbox import operators  # noqa: F401
from pyunlocbox import solvers  # noqa: F401

__version__ = "0.6.1"
__release_date__ = "2025-09-11"


def test():  # pragma: no cover
    """Run the test suite."""
    import unittest

    # Lazy as it might be slow and require additional dependencies.
    from pyunlocbox.tests import suite

    unittest.TextTestRunner(verbosity=2).run(suite)



r"""
The :mod:`pyunlocbox.functions` module implements an interface for solvers to
access the functions to be optimized as well as common objective functions.

Interface
---------

The :class:`func` base class defines a common interface to all functions:

.. autosummary::

    func.cap
    func.eval
    func.prox
    func.grad

Functions
---------

Then, derived classes implement various common objective functions.

**Norm operators** (based on :class:`norm`)

.. autosummary::

    norm_l1
    norm_l2
    norm_nuclear
    norm_tv

**Projection operators** (based on :class:`proj`)

.. autosummary::

    proj_positive
    proj_b2
    proj_lineq
    proj_spsd

**Miscellaneous**

.. autosummary::

    dummy
    structured_sparsity

.. inheritance-diagram:: pyunlocbox.functions
    :parts: 2

"""

import warnings
from copy import deepcopy
from time import time

import numpy as np
from scipy.optimize import minimize

from pyunlocbox import operators as op


def _soft_threshold(z, T, handle_complex=True):
    r"""
    Return the soft thresholded signal.

    Parameters
    ----------
    z : array_like
        Input signal (real or complex).
    T : float or array_like
        Threshold on the absolute value of `z`. There could be either a single
        threshold for the entire signal `z` or one threshold per dimension.
        Useful when you use weighted norms.
    handle_complex : bool
        Indicate that we should handle the thresholding of complex numbers,
        which may be slower. Default is True.

    Returns
    -------
    sz : ndarray
        Soft thresholded signal.

    Examples
    --------
    >>> from pyunlocbox import functions
    >>> functions._soft_threshold([-2, -1, 0, 1, 2], 1)
    array([-1,  0,  0,  0,  1])

    """
    sz = np.maximum(np.abs(z) - T, 0)

    if not handle_complex:
        # This soft thresholding method only supports real signal.
        sz[:] = np.sign(z) * sz

    else:
        # This soft thresholding method supports complex complex signal.
        # Transform to float to avoid integer division.
        # In our case 0 divided by 0 should be 0, not NaN, and is not an error.
        # It corresponds to 0 thresholded by 0, which is 0.
        old_err_state = np.seterr(invalid="ignore")
        sz[:] = np.nan_to_num(1.0 * sz / (sz + T) * z)
        np.seterr(**old_err_state)

    return sz


def _prox_star(func, z, T):
    r"""
    Proximity operator of the convex conjugate of a function.

    Notes
    -----
    Based on the Moreau decomposition of a vector w.r.t. a convex function.

    """
    return z - T * func.prox(z / T, 1 / T)


class func:
    r"""
    This class defines the function object interface.

    It is intended to be a base class for standard functions which will
    implement the required methods. It can also be instantiated by user code
    and dynamically modified for rapid testing.  The instanced objects are
    meant to be passed to the :func:`pyunlocbox.solvers.solve` solving
    function.

    Parameters
    ----------
    y : array_like, optional
        Measurements. Default is 0.
    A : function or ndarray, optional
        The forward operator. Default is the identity, :math:`A(x)=x`. If `A`
        is an ``ndarray``, it will be converted to the operator form.
    At : function or ndarray, optional
        The adjoint operator. If `At` is an ``ndarray``, it will be converted
        to the operator form. If `A` is an ``ndarray``, default is the
        transpose of `A`.  If `A` is a function, default is `A`,
        :math:`At(x)=A(x)`.
    tight : bool, optional
        ``True`` if `A` is a tight frame (semi-orthogonal linear transform),
        ``False`` otherwise. Default is ``True``.
    nu : float, optional
        Bound on the norm of the operator `A`, i.e. :math:`\|A(x)\|^2 \leq \nu
        \|x\|^2`. Default is 1.
    tol : float, optional
        The tolerance stopping criterion. The exact definition depends on the
        function object, please see the documentation of the considered
        function. Default is 1e-3.
    maxit : int, optional
        The maximum number of iterations. Default is 200.

    Examples
    --------

    Let's define a parabola as an example of the manual implementation of a
    function object :

    >>> from pyunlocbox import functions
    >>> f = functions.func()
    >>> f._eval = lambda x: x**2
    >>> f._grad = lambda x: 2*x
    >>> x = [1, 2, 3, 4]
    >>> f.eval(x)
    array([ 1,  4,  9, 16])
    >>> f.grad(x)
    array([2, 4, 6, 8])
    >>> f.cap(x)
    ['EVAL', 'GRAD']

    """

    def __init__(
        self, y=0, A=None, At=None, tight=True, nu=1, tol=1e-3, maxit=200, **kwargs
    ):

        if callable(y):
            self.y = lambda: np.asarray(y())
        else:
            self.y = lambda: np.asarray(y)

        if A is None:
            self.A = lambda x: x
        else:
            if callable(A):
                self.A = A
            else:
                # Transform matrix form to operator form.
                self.A = lambda x: A.dot(x)

        if At is None:
            if A is None:
                self.At = lambda x: x
            elif callable(A):
                self.At = A
            else:
                self.At = lambda x: A.T.dot(x)
        else:
            if callable(At):
                self.At = At
            else:
                self.At = lambda x: At.dot(x)

        self.tight = tight
        self.nu = nu
        self.tol = tol
        self.maxit = maxit

        # Should be initialized if called alone, updated by solve().
        self.verbosity = "NONE"

    def eval(self, x):
        r"""
        Function evaluation.

        Parameters
        ----------
        x : array_like
            The evaluation point. If `x` is a matrix, the function gets
            evaluated for each column, as if it was a set of independent
            problems. Some functions, like the nuclear norm, are only defined
            on matrices.

        Returns
        -------
        z : float
            The objective function evaluated at `x`. If `x` is a matrix, the
            sum of the objectives is returned.

        Notes
        -----
        This method is required by the :func:`pyunlocbox.solvers.solve` solving
        function to evaluate the objective function. Each function class
        should therefore define it.

        """
        sol = self._eval(np.asarray(x))
        if self.verbosity in ["LOW", "HIGH"]:
            name = self.__class__.__name__
            print(f"    {name} evaluation: {sol:e}")
        return sol

    def _eval(self, x):
        raise NotImplementedError("Class user should define this method.")

    def prox(self, x, T):
        r"""
        Function proximal operator.

        Parameters
        ----------
        x : array_like
            The evaluation point. If `x` is a matrix, the function gets
            evaluated for each column, as if it was a set of independent
            problems. Some functions, like the nuclear norm, are only defined
            on matrices.
        T : float
            The regularization parameter.

        Returns
        -------
        z : ndarray
            The proximal operator evaluated for each column of `x`.

        Notes
        -----
        The proximal operator is defined by
        :math:`\operatorname{prox}_{\gamma f}(x) = \operatorname{arg\,min}
        \limits_z \frac{1}{2} \|x-z\|_2^2 + \gamma f(z)`

        This method is required by some solvers.

        When the map A in the function construction is a tight frame
        (semi-orthogonal linear transformation), we can use property (x) of
        Table 10.1 in :cite:`combettes:2011iq` to compute the proximal
        operator of the composition of A with the base function. Whenever
        this is not the case, we have to resort to some iterative procedure,
        which may be very inefficient.

        """
        return self._prox(np.asarray(x), T)

    def _prox(self, x, T):
        raise NotImplementedError("Class user should define this method.")

    def grad(self, x):
        r"""
        Function gradient.

        Parameters
        ----------
        x : array_like
            The evaluation point. If `x` is a matrix, the function gets
            evaluated for each column, as if it was a set of independent
            problems. Some functions, like the nuclear norm, are only defined
            on matrices.

        Returns
        -------
        z : ndarray
            The objective function gradient evaluated for each column of `x`.

        Notes
        -----
        This method is required by some solvers.

        """
        return self._grad(np.asarray(x))

    def _grad(self, x):
        raise NotImplementedError("Class user should define this method.")

    def cap(self, x):
        r"""
        Test the capabilities of the function object.

        Parameters
        ----------
        x : array_like
            The evaluation point. Not really needed, but this function calls
            the methods of the object to test if they can properly execute
            without raising an exception. Therefore it needs some evaluation
            point with a consistent size.

        Returns
        -------
        cap : list of string
            A list of capabilities ('EVAL', 'GRAD', 'PROX').

        """
        tmp = self.verbosity
        self.verbosity = "NONE"
        cap = ["EVAL", "GRAD", "PROX"]
        try:
            self.eval(x)
        except NotImplementedError:
            cap.remove("EVAL")
        try:
            self.grad(x)
        except NotImplementedError:
            cap.remove("GRAD")
        try:
            self.prox(x, 1)
        except NotImplementedError:
            cap.remove("PROX")
        self.verbosity = tmp
        return cap


class dummy(func):
    r"""
    Dummy function (eval, prox, grad).

    This can be used as a second function object when there is only one
    function to minimize. It always evaluates as 0.

    Examples
    --------
    >>> from pyunlocbox import functions
    >>> f = functions.dummy()
    >>> x = [1, 2, 3, 4]
    >>> f.eval(x)
    0
    >>> f.prox(x, 1)
    array([1, 2, 3, 4])
    >>> f.grad(x)
    array([0, 0, 0, 0])

    """

    def __init__(self, **kwargs):
        # Constructor takes keyword-only parameters to prevent user errors.
        super().__init__(**kwargs)

    def _eval(self, x):
        return 0

    def _prox(self, x, T):
        return x

    def _grad(self, x):
        return np.zeros_like(x)


class norm(func):
    r"""
    Base class which defines the attributes of the `norm` objects.

    See generic attributes descriptions of the
    :class:`pyunlocbox.functions.func` base class.

    Parameters
    ----------
    lambda_ : float, optional
        Regularization parameter :math:`\lambda`. Default is 1.
    w : array_like, optional
        Weights for a weighted norm. Default is 1.

    """

    def __init__(self, lambda_=1, w=1, **kwargs):
        super().__init__(**kwargs)
        self.lambda_ = lambda_
        self.w = np.asarray(w)


class norm_l1(norm):
    r"""
    L1-norm (eval, prox).

    See generic attributes descriptions of the
    :class:`pyunlocbox.functions.norm` base class. Note that the constructor
    takes keyword-only parameters.

    Notes
    -----
    * The L1-norm of the vector `x` is given by
      :math:`\lambda \|w \cdot (A(x)-y)\|_1`.
    * The L1-norm proximal operator evaluated at `x` is given by
      :math:`\operatorname{arg\,min}\limits_z \frac{1}{2} \|x-z\|_2^2 + \gamma
      \|w \cdot (A(z)-y)\|_1` where :math:`\gamma = \lambda \cdot T`. This is
      simply a soft thresholding.

    Examples
    --------
    >>> from pyunlocbox import functions
    >>> f = functions.norm_l1()
    >>> f.eval([1, 2, 3, 4])
    10
    >>> f.prox([1, 2, 3, 4], 1)
    array([0., 1., 2., 3.])

    """

    def __init__(self, **kwargs):
        # Constructor takes keyword-only parameters to prevent user errors.
        super().__init__(**kwargs)

    def _eval(self, x):
        sol = self.A(x) - self.y()
        return self.lambda_ * np.sum(np.abs(self.w * sol))

    def _prox(self, x, T):
        # Gamma is T in the matlab UNLocBox implementation.
        gamma = self.lambda_ * T
        if self.tight:
            # Nati: I've checked this code the use of 'y' seems correct
            sol = self.A(x) - self.y()
            sol = _soft_threshold(sol, gamma * self.nu * self.w) - sol
            sol = x + self.At(sol) / self.nu
        else:
            raise NotImplementedError("Not implemented for non-tight frame.")
        return sol


class norm_l2(norm):
    r"""
    L2-norm (eval, prox, grad).

    See generic attributes descriptions of the
    :class:`pyunlocbox.functions.norm` base class. Note that the constructor
    takes keyword-only parameters.

    Notes
    -----
    * The squared L2-norm of the vector `x` is given by
      :math:`\lambda \|w \cdot (A(x)-y)\|_2^2`.
    * The squared L2-norm proximal operator evaluated at `x` is given by
      :math:`\operatorname{arg\,min}\limits_z \frac{1}{2} \|x-z\|_2^2 + \gamma
      \|w \cdot (A(z)-y)\|_2^2` where :math:`\gamma = \lambda \cdot T`.
    * The squared L2-norm gradient evaluated at `x` is given by
      :math:`2 \lambda \cdot At(w \cdot (A(x)-y))`.

    Examples
    --------
    >>> from pyunlocbox import functions
    >>> f = functions.norm_l2()
    >>> x = [1, 2, 3, 4]
    >>> f.eval(x)
    30
    >>> f.prox(x, 1)
    array([0.33333333, 0.66666667, 1.        , 1.33333333])
    >>> f.grad(x)
    array([2, 4, 6, 8])

    """

    def __init__(self, **kwargs):
        # Constructor takes keyword-only parameters to prevent user errors.
        super().__init__(**kwargs)
        if self.tight:
            if self.w.size > 1:
                raise ValueError(
                    "Cannot use tight with vector of weights. Please use a "
                    "constant weight or set tight=False. Alternatively, you "
                    "can merge the weights into your operator."
                )

    def _eval(self, x):
        sol = self.A(x) - self.y()
        return self.lambda_ * np.sum((self.w * sol) ** 2)

    def _prox(self, x, T):
        # Gamma is T in the matlab UNLocBox implementation.
        gamma = self.lambda_ * T
        if self.tight:
            sol = x + 2.0 * gamma * self.At(self.y() * self.w**2)
            sol /= 1.0 + 2.0 * gamma * self.nu * self.w**2
        else:
            res = minimize(
                fun=lambda z: 0.5 * np.sum((z - x) ** 2)
                + gamma * np.sum((self.w * (self.A(z) - self.y())) ** 2),
                x0=x,
                method="BFGS",
                jac=lambda z: z
                - x
                + 2.0 * gamma * self.At((self.w**2) * (self.A(z) - self.y())),
            )
            if res.success:
                sol = res.x
            else:
                raise RuntimeError("norm_l2.prox: " + res.message)
        return sol

    def _grad(self, x):
        sol = self.A(x) - self.y()
        return 2 * self.lambda_ * self.At((self.w**2) * sol)


class norm_nuclear(norm):
    r"""
    Nuclear-norm (eval, prox).

    See generic attributes descriptions of the
    :class:`pyunlocbox.functions.norm` base class. Note that the constructor
    takes keyword-only parameters.

    Notes
    -----
    * The nuclear-norm of the matrix `x` is given by
      :math:`\lambda \| x \|_* = \lambda \operatorname{trace} (\sqrt{x^* x}) =
      \lambda \sum_{i=1}^N |e_i|` where `e_i` are the eigenvalues of `x`.
    * The nuclear-norm proximal operator evaluated at `x` is given by
      :math:`\operatorname{arg\,min}\limits_z \frac{1}{2} \|x-z\|_2^2 + \gamma
      \| x \|_*` where :math:`\gamma = \lambda \cdot T`, which is a
      soft-thresholding of the eigenvalues.

    Examples
    --------
    >>> from pyunlocbox import functions
    >>> f = functions.norm_nuclear()
    >>> f.eval([[1, 2],[2, 3]])  # doctest:+ELLIPSIS
    4.47213595...
    >>> f.prox([[1, 2],[2, 3]], 1)
    array([[0.89442719, 1.4472136 ],
           [1.4472136 , 2.34164079]])

    """

    def __init__(self, is_hermitian=False, **kwargs):
        # Constructor takes keyword-only parameters to prevent user errors.
        super().__init__(**kwargs)
        self.is_hermitian = is_hermitian

    def _eval(self, x):
        # TODO: take care of sparse matrices.
        s = np.linalg.svd(x, compute_uv=False, hermitian=self.is_hermitian)
        return self.lambda_ * np.sum(np.abs(s))

    def _prox(self, x, T):
        # Gamma is T in the matlab UNLocBox implementation.
        gamma = self.lambda_ * T
        # TODO: take care of sparse matrices.
        U, s, V = np.linalg.svd(x, hermitian=self.is_hermitian)
        s = _soft_threshold(s, gamma)
        S = np.diag(s)
        return np.dot(U, np.dot(S, V))


class norm_tv(norm):
    r"""
    TV-norm (eval, prox).

    See generic attributes descriptions of the
    :class:`pyunlocbox.functions.norm` base class. Note that the constructor
    takes keyword-only parameters.

    Notes
    -----
    TODO

    See :cite:`beck2009fastTV` for details about the algorithm.

    Examples
    --------
    >>> import numpy as np
    >>> from pyunlocbox import functions
    >>> f = functions.norm_tv()
    >>> x = np.arange(0, 16)
    >>> x = x.reshape(4, 4)
    >>> f.eval(x)  # doctest:+ELLIPSIS
        norm_tv evaluation: 5.210795e+01
    52.10795063...

    """

    def __init__(self, dim=2, verbosity="LOW", **kwargs):
        super().__init__(**kwargs)
        self.kwargs = kwargs
        self.dim = dim
        self.verbosity = verbosity

    def _eval(self, x):
        if self.dim >= 2:
            y = 0
            grads = []
            grads = op.grad(x, dim=self.dim, **self.kwargs)
            for g in grads:
                y += np.power(abs(g), 2)
            y = np.sqrt(y)
            return np.sum(y)

        if self.dim == 1:
            dx = op.grad(x, dim=self.dim, **self.kwargs)
            y = np.sum(np.abs(dx), axis=0)
            return np.sum(y)

    def _prox(self, x, T):
        # Time counter
        t_init = time()

        tol = self.tol
        maxit = self.maxit

        # TODO implement test_gamma
        # Initialization
        sol = x

        if self.dim == 1:
            r = op.grad(x * 0, dim=self.dim, **self.kwargs)
            rr = deepcopy(r)
        elif self.dim == 2:
            r, s = op.grad(x * 0, dim=self.dim, **self.kwargs)
            rr, ss = deepcopy(r), deepcopy(s)
        elif self.dim == 3:
            r, s, k = op.grad(x * 0, dim=self.dim, **self.kwargs)
            rr, ss, kk = deepcopy(r), deepcopy(s), deepcopy(k)
        elif self.dim == 4:
            r, s, k, u = op.grad(x * 0, dim=self.dim, **self.kwargs)
            rr, ss, kk, uu = deepcopy(r), deepcopy(s), deepcopy(k), deepcopy(u)

        if self.dim >= 1:
            pold = r
        if self.dim >= 2:
            qold = s
        if self.dim >= 3:
            kold = k
        if self.dim >= 4:
            uold = u

        told, prev_obj = 1.0, 0.0

        # Initialization for weights
        if self.dim >= 1:
            try:
                wx = self.kwargs["wx"]
            except (KeyError, TypeError):
                wx = 1.0
        if self.dim >= 2:
            try:
                wy = self.kwargs["wy"]
            except (KeyError, TypeError):
                wy = 1.0
        if self.dim >= 3:
            try:
                wz = self.kwargs["wz"]
            except (KeyError, TypeError):
                wz = 1.0
        if self.dim >= 4:
            try:
                wt = self.kwargs["wt"]
            except (KeyError, TypeError):
                wt = 1.0

        if self.dim == 1:
            mt = wx
        elif self.dim == 2:
            mt = np.maximum(wx, wy)
        elif self.dim == 3:
            mt = np.maximum(wx, np.maximum(wy, wz))
        elif self.dim == 4:
            mt = np.maximum(np.maximum(wx, wy), np.maximum(wz, wt))

        if self.verbosity in ["LOW", "HIGH", "ALL"]:
            print("Proximal TV Operator")

        iter = 0
        crit = "MAX_IT"

        while iter <= maxit:
            # Current Solution
            if self.dim == 1:
                sol = x - T * op.div(rr, **self.kwargs)
            elif self.dim == 2:
                sol = x - T * op.div(rr, ss, **self.kwargs)
            elif self.dim == 3:
                sol = x - T * op.div(rr, ss, kk, **self.kwargs)
            elif self.dim == 4:
                sol = x - T * op.div(rr, ss, kk, uu, **self.kwargs)

            #  Objective function value
            obj = 0.5 * np.power(np.linalg.norm(x[:] - sol[:]), 2) + T * np.sum(
                self._eval(sol), axis=0
            )
            rel_obj = np.abs(obj - prev_obj) / obj
            prev_obj = obj

            if self.verbosity in ["HIGH", "ALL"]:
                print("Iter: ", iter, " obj = ", obj, " rel_obj = ", rel_obj)

            # Stopping criterion
            if rel_obj < tol:
                crit = "TOL_EPS"
                break

            #  Update divergence vectors and project
            if self.dim == 1:
                dx = op.grad(sol, dim=self.dim, **self.kwargs)
                r -= 1.0 / (4 * T * mt**2) * dx
                weights = np.maximum(1, np.abs(r))

            elif self.dim == 2:
                dx, dy = op.grad(sol, dim=self.dim, **self.kwargs)
                r -= (1.0 / (8.0 * T * mt**2.0)) * dx
                s -= (1.0 / (8.0 * T * mt**2.0)) * dy
                weights = np.maximum(
                    1, np.sqrt(np.power(np.abs(r), 2) + np.power(np.abs(s), 2))
                )

            elif self.dim == 3:
                dx, dy, dz = op.grad(sol, dim=self.dim, **self.kwargs)
                r -= 1.0 / (12.0 * T * mt**2) * dx
                s -= 1.0 / (12.0 * T * mt**2) * dy
                k -= 1.0 / (12.0 * T * mt**2) * dz
                weights = np.maximum(
                    1,
                    np.sqrt(
                        np.power(np.abs(r), 2)
                        + np.power(np.abs(s), 2)
                        + np.power(np.abs(k), 2)
                    ),
                )

            elif self.dim == 4:
                dx, dy, dz, dt = op.grad(sol, dim=self.dim, **self.kwargs)
                r -= 1.0 / (16 * T * mt**2) * dx
                s -= 1.0 / (16 * T * mt**2) * dy
                k -= 1.0 / (16 * T * mt**2) * dz
                u -= 1.0 / (16 * T * mt**2) * dt
                weights = np.maximum(
                    1,
                    np.sqrt(
                        np.power(np.abs(r), 2)
                        + np.power(np.abs(s), 2)
                        + np.power(np.abs(k), 2)
                        + np.power(np.abs(u), 2)
                    ),
                )

            # FISTA update
            t = (1 + np.sqrt(4 * told**2)) / 2.0

            if self.dim >= 1:
                p = r / weights
                r = p + (told - 1) / t * (p - pold)
                pold = p
                rr = deepcopy(r)

            if self.dim >= 2:
                q = s / weights
                s = q + (told - 1) / t * (q - qold)
                ss = deepcopy(s)
                qold = q

            if self.dim >= 3:
                o = k / weights
                k = o + (told - 1) / t * (o - kold)
                kk = deepcopy(k)
                kold = o

            if self.dim >= 4:
                m = u / weights
                u = m + (told - 1) / t * (m - uold)
                uu = deepcopy(u)
                uold = m

            told = t
            iter += 1

        t_end = time()
        exec_time = t_end - t_init

        if self.verbosity in ["HIGH", "ALL"]:
            print(
                "Prox_TV: obj = {}, rel_obj = {}, {}, iter = {}".format(
                    obj, rel_obj, crit, iter
                )
            )
            print("exec_time = ", exec_time)
        return sol


class proj(func):
    r"""
    Base class which defines the attributes of the `proj` objects.

    See generic attributes descriptions of the
    :class:`pyunlocbox.functions.func` base class.

    Notes
    -----
    * All indicator functions (projections) evaluate to zero by definition.

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _eval(self, x):
        # Matlab version returns a small delta to avoid division by 0 when
        # evaluating relative tolerance. Here the delta is added in the solve
        # function if the sum of the objective functions is zero.
        # np.spacing(1.0) is equivalent to matlab eps = eps(1.0)
        # return np.spacing(1.0)
        return 0


class proj_positive(proj):
    r"""
    Projection on the positive octant (eval, prox).

    This function is the indicator function :math:`i_S(z)` of the set
    :math:`S = \left\{z \in \mathbb{R}^N \mid z \leq 0 \right\}`
    that is zero if :math:`z` is in the set and infinite otherwise.

    See generic attributes descriptions of the
    :class:`pyunlocbox.functions.proj` base class. Note that the constructor
    takes keyword-only parameters.

    Notes
    -----
    * The evaluation of this function is zero.

    Examples
    --------
    >>> from pyunlocbox import functions
    >>> f = functions.proj_positive()
    >>> x = [-2.5, 1.5]
    >>> f.eval(x)
    0
    >>> f.prox(x, 0)
    array([0. , 1.5])

    """

    def __init__(self, **kwargs):
        # Constructor takes keyword-only parameters to prevent user errors.
        super().__init__(**kwargs)

    def _prox(self, x, T):
        return np.clip(x, 0, np.inf)


class proj_spsd(proj):
    r"""
    Projection on symmetric positive semi-definite matrices (eval, prox).

    This function is the indicator function :math:`i_S(M)` of the set
    :math:`S = \left\{M \in \mathbb{R}^{N \times N}
    \mid M \succeq 0, M=M^T \right\}`
    that is zero if :math:`M` is in the set and infinite otherwise.

    See generic attributes descriptions of the
    :class:`pyunlocbox.functions.proj` base class. Note that the constructor
    takes keyword-only parameters.

    Notes
    -----
    * The evaluation of this function is zero.

    Examples
    --------
    >>> from pyunlocbox import functions
    >>> f = functions.proj_spsd()
    >>> A = np.array([[0, -1] , [-1, 1]])
    >>> A = (A + A.T) / 2  # Symmetrize the matrix.
    >>> np.linalg.eig(A)[0]
    array([-0.61803399,  1.61803399])
    >>> f.eval(A)
    0
    >>> Aproj = f.prox(A, 0)
    >>> np.linalg.eig(Aproj)[0]
    array([0.        , 1.61803399])

    """

    def __init__(self, **kwargs):
        # Constructor takes keyword-only parameters to prevent user errors.
        super().__init__(**kwargs)

    def _prox(self, x, T):
        isreal = np.isreal(x).all()

        # 1. make it symmetric.
        sol = (x + np.conj(x.T)) / 2

        # 2. make it semi-positive.
        D, V = np.linalg.eig(sol)
        D = np.real(D)
        if isreal:
            V = np.real(V)
        D = np.clip(D, 0, np.inf)
        sol = V @ np.diag(D) @ np.conj(V.T)
        return sol


class proj_b2(proj):
    r"""
    Projection on the L2-ball (eval, prox).

    This function is the indicator function :math:`i_S(z)` of the set
    :math:`S= \left\{z \in \mathbb{R}^N \mid \|Az-y\|_2 \leq \epsilon \right\}`
    that is zero if :math:`z` is in the set and infinite otherwise.

    See generic attributes descriptions of the
    :class:`pyunlocbox.functions.proj` base class. Note that the constructor
    takes keyword-only parameters.

    Parameters
    ----------
    epsilon : float, optional
        The radius of the ball. Default is 1.
    method : {'FISTA', 'ISTA'}, optional
        The method used to solve the problem. It can be 'FISTA' or 'ISTA'.
        Default is 'FISTA'.

    Notes
    -----
    * The `tol` parameter is defined as the tolerance for the projection on the
      L2-ball. The algorithm stops if :math:`\frac{\epsilon}{1-tol} \leq
      \|y-A(z)\|_2 \leq \frac{\epsilon}{1+tol}`.
    * The evaluation of this function is zero.
    * The L2-ball proximal operator evaluated at `x` is given by
      :math:`\operatorname{arg\,min}\limits_z \frac{1}{2} \|x-z\|_2^2 + i_S(z)`
      which has an identical solution as
      :math:`\operatorname{arg\,min}\limits_z \|x-z\|_2^2` such that
      :math:`\|A(z)-y\|_2 \leq \epsilon`. It is thus a projection of the vector
      `x` onto an L2-ball of diameter `epsilon`.

    See Also
    --------
    proj_lineq : use instead of ``epsilon=0``

    Examples
    --------
    >>> from pyunlocbox import functions
    >>> f = functions.proj_b2(y=[1, 1])
    >>> x = [3, 3]
    >>> f.eval(x)
    0
    >>> f.prox(x, 0)
    array([1.70710678, 1.70710678])

    """

    def __init__(self, epsilon=1, method="FISTA", **kwargs):
        # Constructor takes keyword-only parameters to prevent user errors.
        super().__init__(**kwargs)
        self.epsilon = epsilon
        self.method = method

    def _prox(self, x, T):

        crit = None  # Stopping criterion.
        niter = 0  # Number of iterations.

        # Tight frame.
        if self.tight:
            tmp1 = self.A(x) - self.y()
            with np.errstate(divide="ignore", invalid="ignore"):
                # Avoid 'division by zero' warning
                scale = self.epsilon / np.sqrt(np.sum(tmp1 * tmp1, axis=0))
            tmp2 = tmp1 * np.minimum(1, scale)  # Scaling.
            sol = x + self.At(tmp2 - tmp1) / self.nu
            crit = "TOL"
            u = np.nan

        # Non tight frame.
        else:

            # Initialization.
            sol = x
            u = np.zeros_like(self.y())
            if self.method == "FISTA":
                v_last = u
                t_last = 1.0
            elif self.method != "ISTA":
                raise ValueError("The method should be either FISTA or ISTA.")

            # Tolerance around the L2-ball.
            epsilon_low = self.epsilon / (1.0 + self.tol)
            epsilon_up = self.epsilon / (1.0 - self.tol)

            # Check if we are already in the L2-ball.
            norm_res = np.linalg.norm(self.y() - self.A(sol), 2)
            if norm_res <= epsilon_up:
                crit = "INBALL"

            # Projection onto the L2-ball
            while not crit:

                niter += 1

                # Residual.
                res = self.A(sol) - self.y()
                norm_res = np.linalg.norm(res, 2)

                if self.verbosity == "HIGH":
                    print(
                        "    proj_b2 iteration {:3d}: epsilon = {:.2e}, "
                        "||y-A(z)||_2 = {:.2e}".format(niter, self.epsilon, norm_res)
                    )

                # Scaling for projection.
                res += u * self.nu
                norm_proj = np.linalg.norm(res, 2)

                ratio = min(1, self.epsilon / norm_proj)
                v = 1.0 / self.nu * (res - res * ratio)

                if self.method == "FISTA":
                    t = (1.0 + np.sqrt(1.0 + 4.0 * t_last**2.0)) / 2.0  # Time step.
                    u = v + (t_last - 1.0) / t * (v - v_last)
                    v_last = v
                    t_last = t
                else:
                    u = v

                # Current estimation.
                sol = x - self.At(u)

                # Stopping criterion.
                if norm_res >= epsilon_low and norm_res <= epsilon_up:
                    crit = "TOL"
                elif niter >= self.maxit:
                    crit = "MAXIT"

            if self.verbosity in ["LOW", "HIGH"]:
                norm_res = np.linalg.norm(self.y() - self.A(sol), 2)
                print(
                    "    proj_b2: epsilon = {:.2e}, ||y-A(z)||_2 = {:.2e}, "
                    "{}, niter = {}".format(self.epsilon, norm_res, crit, niter)
                )

        return sol


class proj_lineq(proj):
    r"""
    Projection on the plane satisfying the linear equality Az = y (eval, prox).

    This function is the indicator function :math:`i_S(z)` of the set
    :math:`S = \left\{z \in \mathbb{R}^N \mid Az = y \right\}`
    that is zero if :math:`z` is in the set and infinite otherwise.

    The proximal operator is
    :math:`\operatorname{arg\,min}_z \| z - x \|_2 \text{ s.t. } Az = y`.

    See generic attributes descriptions of the
    :class:`pyunlocbox.functions.proj` base class. Note that the constructor
    takes keyword-only parameters.

    Notes
    -----
    * A parameter `pinvA`, the pseudo-inverse of `A`, must be provided if the
      parameter `A` is provided as an operator/callable (not a matrix).
    * The evaluation of this function is zero.

    See Also
    --------
    proj_b2 : quadratic case

    Examples
    --------
    >>> from pyunlocbox import functions
    >>> import numpy as np
    >>> x = np.array([0, 0])
    >>> A = np.array([[1, 1]])
    >>> pinvA = np.linalg.pinv(A)
    >>> y = np.array([1])
    >>> f = functions.proj_lineq(A=A, pinvA=pinvA, y=y)
    >>> sol = f.prox(x, 0)
    >>> sol
    array([0.5, 0.5])
    >>> np.abs(A.dot(sol) - y) < 1e-15
    array([ True])

    """

    def __init__(self, A=None, pinvA=None, **kwargs):
        # Constructor takes keyword-only parameters to prevent user errors.
        super().__init__(A=A, **kwargs)

        if pinvA is None:
            if A is None:
                warnings.warn(
                    "Are you sure about the parameters? "
                    "The projection will return y.",
                    RuntimeWarning,
                )
                self.pinvA = lambda x: x
            else:
                if callable(A):
                    raise ValueError("Provide A as a numpy array or provide pinvA.")
                else:
                    # Transform matrix form to operator form.
                    self._pinvA = np.linalg.pinv(A)
                    self.pinvA = lambda x: self._pinvA.dot(x)
        else:
            if callable(pinvA):
                self.pinvA = pinvA
            else:
                self.pinvA = lambda x: pinvA.dot(x)

    def _prox(self, x, T):
        # Applying the projection formula.
        # (for now, only the non scalable version)
        residue = self.A(x) - self.y()
        sol = x - self.pinvA(residue)
        return sol


class structured_sparsity(func):
    r"""
    Structured sparsity (eval, prox).

    The structured sparsity term that is defined in the work of
    Jenatton et al. 2011 `Proximal methods for hierarchical sparse coding
    <https://hal.inria.fr/inria-00516723>`_.

    .. math:: \Omega(x) = \lambda \cdot \sum_{g \in G} w_g \cdot \|x_g\|_2

    See generic attributes descriptions of the
    :class:`pyunlocbox.functions.func` base class.

    Parameters
    ----------
    lambda_ : float, optional
        The scaling factor of the function that corresponds to :math:`\lambda`.
        Must be a non-negative number.
    groups: list of lists of integers
        Each element encodes the indices of the vector belonging to a single
        group. Corresponds to :math:`G`.
    weights : array_like
        Weight associated to each group. Corresponds to :math:`w_g`. Must have
        the same length as :math:`G`.

    Examples
    --------
    >>> from pyunlocbox import functions
    >>> groups = [[0, 1], [3, 2, 4]]
    >>> weights = [2, 1]
    >>> f = functions.structured_sparsity(10, groups, weights)
    >>> x = [2, 2.5, -0.5, 0.3, 0.01]
    >>> f.eval(x)
    69.86305169905782
    >>> f.prox(x, 0.1)
    array([0.7506099 , 0.93826238, 0.        , 0.        , 0.        ])

    """

    def __init__(self, lambda_=1, groups=[[]], weights=[0], **kwargs):
        super().__init__(**kwargs)

        if lambda_ < 0:
            raise ValueError("The scaling factor must be non-negative.")
        self.lambda_ = lambda_

        if not isinstance(groups, list):
            raise TypeError("The groups must be defined as a list of lists.")
        self.groups = groups

        if len(weights) != len(groups):
            raise ValueError("Length of weights must be equal to number of " "groups.")
        self.weights = weights

    def _eval(self, x):
        costs = [w * np.linalg.norm(x[g]) for g, w in zip(self.groups, self.weights)]
        return self.lambda_ * np.sum(costs)

    def _prox(self, x, T):
        gamma = self.lambda_ * T
        v = x.copy()
        for g, w in zip(self.groups, self.weights):
            xn = np.linalg.norm(v[g])
            r = gamma * w
            if xn > r:
                v[g] -= v[g] * r / xn
            else:
                v[g] = 0
        return v



r"""
The :mod:`pyunlocbox.solvers` module implements a solving function (which will
minimize your objective function) as well as common solvers.

Solving
-------

Call :func:`solve` to solve your convex optimization problem using your
instantiated solver and functions objects.

Interface
---------

The :class:`solver` base class defines a common interface to all solvers:

.. autosummary::

    solver.pre
    solver.algo
    solver.post

Solvers
-------

Then, derived classes implement various common solvers.

.. autosummary::

    gradient_descent
    forward_backward
    douglas_rachford
    generalized_forward_backward

**Primal-dual solvers** (based on :class:`primal_dual`)

.. autosummary::

    chambolle_pock
    mlfbf
    projection_based

.. inheritance-diagram:: pyunlocbox.solvers
    :parts: 2

"""

import time

import numpy as np

from pyunlocbox import acceleration
from pyunlocbox.functions import _prox_star, dummy


def solve(
    functions,
    x0,
    solver=None,
    atol=None,
    dtol=None,
    rtol=1e-3,
    xtol=None,
    maxit=200,
    verbosity="LOW",
    inplace=False,
):
    r"""
    Solve an optimization problem whose objective function is the sum of some
    convex functions.

    This function minimizes the objective function :math:`f(x) =
    \sum\limits_{k=0}^{k=K} f_k(x)`, i.e. solves
    :math:`\operatorname{arg\,min}\limits_x f(x)` for :math:`x \in
    \mathbb{R}^{n \times N}` where :math:`n` is the dimensionality of the data
    and :math:`N` the number of independent problems. It returns a dictionary
    with the found solution and some informations about the algorithm
    execution.

    Parameters
    ----------
    functions : list of objects
        A list of convex functions to minimize. These are objects who must
        implement the :meth:`pyunlocbox.functions.func.eval` method. The
        :meth:`pyunlocbox.functions.func.grad` and / or
        :meth:`pyunlocbox.functions.func.prox` methods are required by some
        solvers. Note also that some solvers can only handle two convex
        functions while others may handle more. Please refer to the
        documentation of the considered solver.
    x0 : array_like
        Starting point of the algorithm, :math:`x_0 \in \mathbb{R}^{n \times
        N}`.
    solver : solver class instance, optional
        The solver algorithm. It is an object who must inherit from
        :class:`pyunlocbox.solvers.solver` and implement the :meth:`_pre`,
        :meth:`_algo` and :meth:`_post` methods. If no solver object are
        provided, a standard one will be chosen given the number of convex
        function objects and their implemented methods.
    atol : float, optional
        The absolute tolerance stopping criterion. The algorithm stops when
        :math:`f(x^t) < atol` where :math:`f(x^t)` is the objective function at
        iteration :math:`t`. Default is None.
    dtol : float, optional
        Stop when the objective function is stable enough, i.e. when
        :math:`\left|f(x^t) - f(x^{t-1})\right| < dtol`. Default is None.
    rtol : float, optional
        The relative tolerance stopping criterion. The algorithm stops when
        :math:`\left|\frac{ f(x^t) - f(x^{t-1}) }{ f(x^t) }\right| < rtol`.
        Default is :math:`10^{-3}`.
    xtol : float, optional
        Stop when the variable is stable enough, i.e. when :math:`\frac{\|x^t -
        x^{t-1}\|_2}{\sqrt{n N}} < xtol`. Note that additional memory will be
        used to store :math:`x^{t-1}`. Default is None.
    maxit : int, optional
        The maximum number of iterations. Default is 200.
    verbosity : {'NONE', 'LOW', 'HIGH', 'ALL'}, optional
        The log level : ``'NONE'`` for no log, ``'LOW'`` for resume at
        convergence, ``'HIGH'`` for info at all solving steps, ``'ALL'`` for
        all possible outputs, including at each steps of the proximal operators
        computation. Default is ``'LOW'``.
    inplace : bool, optional
        If True and x0 is a numpy array, then x0 will be modified in place
        during execution to save memory. It will then contain the solution. Be
        careful to pass data of the type (int, float32, float64) you want your
        computations to use.

    Returns
    -------
    sol : ndarray
        The problem solution.
    solver : str
        The used solver.
    crit : {'ATOL', 'DTOL', 'RTOL', 'XTOL', 'MAXIT'}
        The used stopping criterion. See above for definitions.
    niter : int
        The number of iterations.
    time : float
        The execution time in seconds.
    objective : ndarray
        The successive evaluations of the objective function at each iteration.

    Examples
    --------
    >>> import numpy as np
    >>> from pyunlocbox import functions, solvers

    Define a problem:

    >>> y = [4, 5, 6, 7]
    >>> f = functions.norm_l2(y=y)

    Solve it:

    >>> x0 = np.zeros(len(y))
    >>> ret = solvers.solve([f], x0, atol=1e-2, verbosity='ALL')
    INFO: Dummy objective function added.
    INFO: Selected solver: forward_backward
    INFO: Forward-backward method
        dummy evaluation: 0.000000e+00
        norm_l2 evaluation: 1.260000e+02
    Iteration 1 of forward_backward:
        dummy evaluation: 0.000000e+00
        norm_l2 evaluation: 1.400000e+01
        objective = 1.40e+01
    Iteration 2 of forward_backward:
        dummy evaluation: 0.000000e+00
        norm_l2 evaluation: 2.963739e-01
        objective = 2.96e-01
    Iteration 3 of forward_backward:
        dummy evaluation: 0.000000e+00
        norm_l2 evaluation: 7.902529e-02
        objective = 7.90e-02
    Iteration 4 of forward_backward:
        dummy evaluation: 0.000000e+00
        norm_l2 evaluation: 5.752265e-02
        objective = 5.75e-02
    Iteration 5 of forward_backward:
        dummy evaluation: 0.000000e+00
        norm_l2 evaluation: 5.142032e-03
        objective = 5.14e-03
    Solution found after 5 iterations:
        objective function f(sol) = 5.142032e-03
        stopping criterion: ATOL

    Verify the stopping criterion (should be smaller than atol=1e-2):

    >>> np.linalg.norm(ret['sol'] - y)**2  # doctest:+ELLIPSIS
    0.00514203...

    Show the solution (should be close to y w.r.t. the L2-norm measure):

    >>> ret['sol']
    array([4.02555301, 5.03194126, 6.03832952, 7.04471777])

    Show the used solver:

    >>> ret['solver']
    'forward_backward'

    Show some information about the convergence:

    >>> ret['crit']
    'ATOL'
    >>> ret['niter']
    5
    >>> ret['time']  # doctest:+SKIP
    0.0012578964233398438
    >>> ret['objective']  # doctest:+NORMALIZE_WHITESPACE,+ELLIPSIS
    [[126.0, 0], [13.99999999..., 0], [0.29637392..., 0], [0.07902528..., 0],
    [0.05752265..., 0], [0.00514203..., 0]]

    """
    # to prevent any modification of the input
    if not (inplace):
        x0 = x0.copy()

    if verbosity not in ["NONE", "LOW", "HIGH", "ALL"]:
        raise ValueError("Verbosity should be either NONE, LOW, HIGH or ALL.")

    # Add a second dummy convex function if only one function is provided.
    if len(functions) < 1:
        raise ValueError("At least 1 convex function should be provided.")
    elif len(functions) == 1:
        functions.append(dummy())
        if verbosity in ["LOW", "HIGH", "ALL"]:
            print("INFO: Dummy objective function added.")

    # Choose a solver if none provided.
    if not solver:
        if len(functions) == 2:
            fb0 = "GRAD" in functions[0].cap(x0) and "PROX" in functions[1].cap(x0)
            fb1 = "GRAD" in functions[1].cap(x0) and "PROX" in functions[0].cap(x0)
            dg0 = "PROX" in functions[0].cap(x0) and "PROX" in functions[1].cap(x0)
            if fb0 or fb1:
                solver = forward_backward()  # Need one prox and 1 grad.
            elif dg0:
                solver = douglas_rachford()  # Need two prox.
            else:
                raise ValueError("No suitable solver for the given functions.")
        elif len(functions) > 2:
            solver = generalized_forward_backward()
        if verbosity in ["LOW", "HIGH", "ALL"]:
            name = solver.__class__.__name__
            print(f"INFO: Selected solver: {name}")

    # Set solver and functions verbosity.
    translation = {"ALL": "HIGH", "HIGH": "HIGH", "LOW": "LOW", "NONE": "NONE"}
    solver.verbosity = translation[verbosity]
    translation = {"ALL": "HIGH", "HIGH": "LOW", "LOW": "NONE", "NONE": "NONE"}
    functions_verbosity = []
    for f in functions:
        functions_verbosity.append(f.verbosity)
        f.verbosity = translation[verbosity]

    tstart = time.time()
    crit = None
    niter = 0
    rtol_only_zeros = True

    # Solver specific initialization.
    solver.pre(functions, x0)

    # Evaluate the objective function at the begining
    objective = [solver.objective(x0)]

    while not crit:

        niter += 1

        if xtol is not None:
            last_sol = np.array(solver.sol, copy=True)

        if verbosity in ["HIGH", "ALL"]:
            name = solver.__class__.__name__
            print(f"Iteration {niter} of {name}:")

        # Solver iterative algorithm.
        solver.algo(objective, niter)

        objective.append(solver.objective(solver.sol))
        current = np.sum(objective[-1])
        last = np.sum(objective[-2])

        # Verify stopping criteria.
        if atol is not None and current < atol:
            crit = "ATOL"
        if dtol is not None and np.abs(current - last) < dtol:
            crit = "DTOL"
        if rtol is not None:
            div = current  # Prevent division by 0.
            if div == 0:
                if verbosity in ["LOW", "HIGH", "ALL"]:
                    print("WARNING: (rtol) objective function is equal to 0 !")
                if last != 0:
                    div = last
                else:
                    div = 1.0  # Result will be zero anyway.
            else:
                rtol_only_zeros = False
            relative = np.abs((current - last) / div)
            if relative < rtol and not rtol_only_zeros:
                crit = "RTOL"
        if xtol is not None:
            err = np.linalg.norm(solver.sol - last_sol)
            err /= np.sqrt(last_sol.size)
            if err < xtol:
                crit = "XTOL"
        if maxit is not None and niter >= maxit:
            crit = "MAXIT"

        if verbosity in ["HIGH", "ALL"]:
            print(f"    objective = {current:.2e}")

    # Restore verbosity for functions. In case they are called outside solve().
    for k, f in enumerate(functions):
        f.verbosity = functions_verbosity[k]

    if verbosity in ["LOW", "HIGH", "ALL"]:
        print(f"Solution found after {niter} iterations:")
        print(f"    objective function f(sol) = {current:e}")
        print(f"    stopping criterion: {crit}")

    # Returned dictionary.
    result = {
        "sol": solver.sol,
        "solver": solver.__class__.__name__,  # algo for consistency ?
        "crit": crit,
        "niter": niter,
        "time": time.time() - tstart,
        "objective": objective,
    }
    try:
        # Update dictionary for primal-dual solvers
        result["dual_sol"] = solver.dual_sol
    except AttributeError:
        pass

    # Solver specific post-processing (e.g. delete references).
    solver.post()

    return result


class solver:
    r"""
    Defines the solver object interface.

    This class defines the interface of a solver object intended to be passed
    to the :func:`pyunlocbox.solvers.solve` solving function. It is intended to
    be a base class for standard solvers which will implement the required
    methods. It can also be instantiated by user code and dynamically modified
    for rapid testing. This class also defines the generic attributes of all
    solver objects.

    Parameters
    ----------
    step : float
        The gradient-descent step-size. This parameter is bounded by 0 and
        :math:`\frac{2}{\beta}` where :math:`\beta` is the Lipschitz constant
        of the gradient of the smooth function (or a sum of smooth functions).
        Default is 1.
    accel : pyunlocbox.acceleration.accel
        User-defined object used to adaptively change the current step size
        and solution while the algorithm is running. Default is a dummy
        object that returns unchanged values.

    """

    def __init__(self, step=1.0, accel=None):
        if step < 0:
            raise ValueError("Step should be a positive number.")
        self.step = step
        self.accel = acceleration.dummy() if accel is None else accel

    def pre(self, functions, x0):
        """
        Solver-specific pre-processing. See parameters documentation in
        :func:`pyunlocbox.solvers.solve` documentation.

        Notes
        -----
        When preprocessing the functions, the solver should split them into
        two lists:
        * `self.smooth_funs`, for functions involved in gradient steps.
        * `self.non_smooth_funs`, for functions involved proximal steps.
        This way, any method that takes in the solver as argument, such as the
        methods in :class:`pyunlocbox.acceleration.accel`, can have some
        context as to how the solver is using the functions.

        """
        self.sol = np.asarray(x0)
        self.smooth_funs = []
        self.non_smooth_funs = []
        self._pre(functions, self.sol)
        self.accel.pre(functions, self.sol)

    def _pre(self, functions, x0):
        raise NotImplementedError("Class user should define this method.")

    def algo(self, objective, niter):
        """
        Call the solver iterative algorithm and the provided acceleration
        scheme. See parameters documentation in
        :func:`pyunlocbox.solvers.solve`

        Notes
        -----
        The method :meth:`self.accel.update_sol` is called before
        :meth:`self._algo` because the acceleration schemes usually involves
        some sort of averaging of previous solutions, which can add some
        unwanted artifacts on the output solution. With this ordering, we
        guarantee that the output of solver.algo is not corrupted by the
        acceleration scheme.

        Similarly, the method :meth:`self.accel.update_step` is called after
        :meth:`self._algo` to allow the step update procedure to act directly
        on the solution output by the underlying algorithm, and not on the
        intermediate solution output by the acceleration scheme in
        :meth:`self.accel.update_sol`.

        """
        self.sol[:] = self.accel.update_sol(self, objective, niter)
        self.step = self.accel.update_step(self, objective, niter)
        self._algo()

    def _algo(self):
        raise NotImplementedError("Class user should define this method.")

    def post(self):
        """
        Solver-specific post-processing. Mainly used to delete references added
        during initialization so that the garbage collector can free the
        memory. See parameters documentation in
        :func:`pyunlocbox.solvers.solve`.

        """
        self._post()
        self.accel.post()
        del self.sol, self.smooth_funs, self.non_smooth_funs

    def _post(self):
        raise NotImplementedError("Class user should define this method.")

    def objective(self, x):
        """
        Return the objective function at x.

        Necessitate `solver._pre(...)` to be run first.
        """
        return self._objective(x)

    def _objective(self, x):
        obj_smooth = [f.eval(x) for f in self.smooth_funs]
        obj_nonsmooth = [f.eval(x) for f in self.non_smooth_funs]
        return obj_nonsmooth + obj_smooth


class gradient_descent(solver):
    r"""
    Gradient descent algorithm.

    This algorithm solves optimization problems composed of the sum of
    any number of smooth functions.

    See generic attributes descriptions of the
    :class:`pyunlocbox.solvers.solver` base class.

    Notes
    -----
    This algorithm requires each function implement the
    :meth:`pyunlocbox.functions.func.grad` method.

    See :class:`pyunlocbox.acceleration.regularized_nonlinear` for a very
    efficient acceleration scheme for this method.

    Examples
    --------
    >>> import numpy as np
    >>> from pyunlocbox import functions, solvers
    >>> dim = 25
    >>> np.random.seed(0)
    >>> xstar = np.random.rand(dim)  # True solution
    >>> x0 = np.random.rand(dim)
    >>> x0 = xstar + 5*(x0 - xstar) / np.linalg.norm(x0 - xstar)
    >>> A = np.random.rand(dim, dim)
    >>> step = 1 / np.linalg.norm(np.dot(A.T, A))
    >>> f = functions.norm_l2(lambda_=0.5, A=A, y=np.dot(A, xstar))
    >>> fd = functions.dummy()
    >>> solver = solvers.gradient_descent(step=step)
    >>> params = {'rtol':0, 'maxit':14000, 'verbosity':'NONE'}
    >>> ret = solvers.solve([f, fd], x0, solver, **params)
    >>> pctdiff = 100 * np.sum((xstar - ret['sol'])**2) / np.sum(xstar**2)
    >>> print('Difference: {0:.1f}%'.format(pctdiff))
    Difference: 1.3%

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _pre(self, functions, x0):

        for f in functions:
            if "GRAD" in f.cap(x0):
                self.smooth_funs.append(f)
            else:
                raise ValueError(
                    "Gradient descent requires each function to " "implement grad()."
                )

        if self.verbosity == "HIGH":
            print(
                "INFO: Gradient descent minimizing {} smooth "
                "functions.".format(len(self.smooth_funs))
            )

    def _algo(self):
        """
        x^{k+1} = x^k-λ ∇x^k
        """
        grad = np.zeros_like(self.sol)
        for f in self.smooth_funs:
            grad += f.grad(self.sol)
        self.sol[:] -= self.step * grad

    def _post(self):
        pass


class forward_backward(solver):
    r"""
    Forward-backward proximal splitting (FISTA and ISTA) algorithm.

    This algorithm solves convex optimization problems composed of the sum of
    a smooth and a non-smooth function.

    See generic attributes descriptions of the
    :class:`pyunlocbox.solvers.solver` base class.

    Parameters
    ----------
    accel : :class:`pyunlocbox.acceleration.accel`
        Acceleration scheme to use.
        Default is :meth:`pyunlocbox.acceleration.fista`, which corresponds
        to the 'FISTA' solver. Passing :meth:`pyunlocbox.acceleration.dummy`
        instead results in the ISTA solver. Note that while FISTA is much more
        time-efficient, it is less memory-efficient.

    Notes
    -----
    This algorithm requires one function to implement the
    :meth:`pyunlocbox.functions.func.prox` method and the other one to
    implement the :meth:`pyunlocbox.functions.func.grad` method.

    See :cite:`beck2009FISTA` for details about the algorithm.

    Examples
    --------
    >>> import numpy as np
    >>> from pyunlocbox import functions, solvers
    >>> y = [4, 5, 6, 7]
    >>> x0 = np.zeros(len(y))
    >>> f1 = functions.norm_l2(y=y)
    >>> f2 = functions.dummy()
    >>> solver = solvers.forward_backward(step=0.5)
    >>> ret = solvers.solve([f1, f2], x0, solver, atol=1e-5)
    Solution found after 15 iterations:
        objective function f(sol) = 4.957288e-07
        stopping criterion: ATOL
    >>> ret['sol']
    array([4.0002509 , 5.00031362, 6.00037635, 7.00043907])

    """

    def __init__(self, accel=acceleration.fista(), **kwargs):
        super().__init__(accel=accel, **kwargs)

    def _pre(self, functions, x0):

        if self.verbosity == "HIGH":
            print("INFO: Forward-backward method")

        if len(functions) != 2:
            raise ValueError("Forward-backward requires two convex functions.")

        if "PROX" in functions[0].cap(x0) and "GRAD" in functions[1].cap(x0):
            self.smooth_funs.append(functions[1])
            self.non_smooth_funs.append(functions[0])
        elif "PROX" in functions[1].cap(x0) and "GRAD" in functions[0].cap(x0):
            self.smooth_funs.append(functions[0])
            self.non_smooth_funs.append(functions[1])
        else:
            raise ValueError(
                "Forward-backward requires a function to "
                "implement prox() and the other grad()."
            )

    def _algo(self):
        # Forward step
        x = self.sol - self.step * self.smooth_funs[0].grad(self.sol)
        # Backward step
        self.sol[:] = self.non_smooth_funs[0].prox(x, self.step)

    def _post(self):
        pass


class generalized_forward_backward(solver):
    r"""
    Generalized forward-backward proximal splitting algorithm.

    This algorithm solves convex optimization problems composed of the sum of
    any number of non-smooth (or smooth) functions.

    See generic attributes descriptions of the
    :class:`pyunlocbox.solvers.solver` base class.

    Parameters
    ----------
    lambda_ : float, optional
        A relaxation parameter bounded by 0 and 1. Default is 1.

    Notes
    -----
    This algorithm requires each function to either implement the
    :meth:`pyunlocbox.functions.func.prox` method or the
    :meth:`pyunlocbox.functions.func.grad` method.

    See :cite:`raguet2013generalizedFB` for details about the algorithm.

    Examples
    --------
    >>> import numpy as np
    >>> from pyunlocbox import functions, solvers
    >>> y = [0.01, 0.2, 8, 0.3, 0 , 0.03, 7]
    >>> x0 = np.zeros(len(y))
    >>> f1 = functions.norm_l2(y=y)
    >>> f2 = functions.norm_l1()
    >>> solver = solvers.generalized_forward_backward(lambda_=1, step=0.5)
    >>> ret = solvers.solve([f1, f2], x0, solver)
    Solution found after 2 iterations:
        objective function f(sol) = 1.463100e+01
        stopping criterion: RTOL
    >>> ret['sol']
    array([0. , 0. , 7.5, 0. , 0. , 0. , 6.5])

    """

    def __init__(self, lambda_=1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lambda_ = lambda_

    def _pre(self, functions, x0):

        if self.lambda_ <= 0 or self.lambda_ > 1:
            raise ValueError("Lambda is bounded by 0 and 1.")

        self.z = []
        for f in functions:
            if "GRAD" in f.cap(x0):
                self.smooth_funs.append(f)
            elif "PROX" in f.cap(x0):
                self.non_smooth_funs.append(f)
                self.z.append(np.array(x0, copy=True))
            else:
                raise ValueError(
                    "Generalized forward-backward requires each "
                    "function to implement prox() or grad()."
                )

        if self.verbosity == "HIGH":
            print(
                "INFO: Generalized forward-backward minimizing {} smooth "
                "functions and {} non-smooth functions.".format(
                    len(self.smooth_funs), len(self.non_smooth_funs)
                )
            )

    def _algo(self):

        # Smooth functions.
        grad = np.zeros_like(self.sol)
        for f in self.smooth_funs:
            grad += f.grad(self.sol)

        # Non-smooth functions.
        if not self.non_smooth_funs:
            self.sol[:] -= self.step * grad  # Reduces to gradient descent.
        else:
            sol = np.zeros_like(self.sol)
            for i, g in enumerate(self.non_smooth_funs):
                tmp = 2 * self.sol - self.z[i] - self.step * grad
                tmp[:] = g.prox(tmp, self.step * len(self.non_smooth_funs))
                self.z[i] += self.lambda_ * (tmp - self.sol)
                sol += 1.0 * self.z[i] / len(self.non_smooth_funs)
            self.sol[:] = sol

    def _post(self):
        del self.z


class douglas_rachford(solver):
    r"""
    Douglas-Rachford proximal splitting algorithm.

    This algorithm solves convex optimization problems composed of the sum of
    two non-smooth (or smooth) functions.

    See generic attributes descriptions of the
    :class:`pyunlocbox.solvers.solver` base class.

    Parameters
    ----------
    lambda_ : float, optional
        The update term weight. It should be between 0 and 1. Default is 1.
    A : array_like, optional
        Matrix implementing a linear transformation of x in g() as :
        minimize f(x) + g(Ax)

    Notes
    -----
    This algorithm requires the two functions to implement the
    :meth:`pyunlocbox.functions.func.prox` method.

    See :cite:`combettes2007DR` for details about the algorithm.

    Examples
    --------
    >>> import numpy as np
    >>> from pyunlocbox import functions, solvers
    >>> y = [4, 5, 6, 7]
    >>> x0 = np.zeros(len(y))
    >>> f1 = functions.norm_l2(y=y)
    >>> f2 = functions.dummy()
    >>> solver = solvers.douglas_rachford(lambda_=1, step=1)
    >>> ret = solvers.solve([f1, f2], x0, solver, atol=1e-5)
    Solution found after 8 iterations:
        objective function f(sol) = 2.927052e-06
        stopping criterion: ATOL
    >>> ret['sol']
    array([3.99939034, 4.99923792, 5.99908551, 6.99893309])

    Linearized ADMM:

    >>> import numpy as np
    >>> from pyunlocbox import functions, solvers
    >>> y = np.array([4, -9, -13, -4])
    >>> L = np.array([[5, 9, 3], [7, 8, 5], [4, 4, 9], [0, 1, 7]])
    >>> max_step = 0.5 / (1 + np.linalg.norm(L, 2))
    >>> x0 = np.zeros(3)
    >>> f1 = functions.norm_l1()
    >>> f2 = functions.norm_l1(y=y)
    >>> solver = solvers.douglas_rachford(step=max_step*50, A=L)
    >>> ret = solvers.solve([f1, f2], x0, solver, atol=1e-1, maxit=1000, rtol=1e-5)
    Solution found after 993 iterations:
        objective function f(sol) = 8.008191e+00
        stopping criterion: RTOL
    >>> ret['sol']
    array([-4.00133346,  3.00096956, -0.99996531])

    """

    def __init__(self, lambda_=1, A=None, mu=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lambda_ = lambda_
        self._has_A = A is not None

        if A is None:
            self.A = lambda x: x
            self.At = lambda x: x
        else:
            # Transform matrix form to operator form.
            self.A = lambda x: A.dot(x)
            self.At = lambda x: A.T.dot(x)

        if mu is None:
            if A is not None:
                self.mu = self.step / (np.linalg.norm(A, 2) ** 2)
            else:
                self.mu = 0.5  # Should check this...

    def _pre(self, functions, x0):

        if self.lambda_ <= 0 or self.lambda_ > 1:
            raise ValueError("Lambda is bounded by 0 and 1.")
        if self.mu <= 0 or self.mu > 1:
            raise ValueError("Mu is bounded by 0 and 1.")

        if len(functions) != 2:
            raise ValueError("Douglas-Rachford requires two convex functions.")

        for f in functions:
            x1 = np.copy(x0)

            try:
                f.cap(x1)
            except ValueError:
                x1 = self.A(x0)

            if "PROX" in f.cap(x1):
                self.non_smooth_funs.append(f)
            else:
                raise ValueError(
                    "Douglas-Rachford requires each " "function to implement prox()."
                )

        self.z = np.array(self.A(x0), copy=True)
        self.u = np.array(self.A(x0), copy=True)

    def _algo(self):
        """
        Default:
            x^{k+1} = prox_{θf} (z^k)
            z^{k+1} = z^k + λ (prox_{θg}(2x^{k+1}−z^k) − x^{k+1})
        Or equivalently:
            z^{k+1} = prox_{θg} (x^k + u^k)
            x^{k+1} = prox_{µf} (z^{k+1} − u^k)
            u^{k+1} = u^k + λ (x^{k+1} − z^{k+1})

        If linearized:
            z^{k+1} = prox_{θg} (Ax^k+u^k)
            x^{k+1} = prox_{µf} (x^k − (µ/θ)A^T(Ax^k−z^{k+1}+u^k))
            u^{k+1} = u^k + λ (Ax^{k+1} − z^{k+1})

        """
        if not self._has_A:
            tmp = self.non_smooth_funs[0].prox(2 * self.sol - self.z, self.step)
            self.z[:] = self.z + self.lambda_ * (
                tmp - self.sol
            )  # prox_{λg}(y) != λ prox_{g}(y)
            self.sol[:] = self.non_smooth_funs[1].prox(self.z, self.step)

        #     # self.z[:] = self.non_smooth_funs[0].prox(self.sol + self.u, self.step)
        #     # self.sol[:] = self.non_smooth_funs[1].prox(self.z-self.u, self.step)
        #     # self.u[:] = self.u + self.sol - self.z

        else:
            # See "Proximal Algorithms. N. Parikh and S. Boyd.
            # Foundations and Trends in Optimization, 1(3):123-231, 2014."
            self.z[:] = self.non_smooth_funs[1].prox(
                self.A(self.sol) + self.u, self.step
            )
            self.sol[:] = self.non_smooth_funs[0].prox(
                self.sol
                - (self.mu / self.step) * self.At(self.A(self.sol) - self.z + self.u),
                self.mu,
            )
            self.u[:] = self.u + self.lambda_ * (self.A(self.sol) - self.z)

    def _objective(self, x):
        obj_smooth = [f.eval(x) for f in self.smooth_funs]
        obj_nonsmooth = [
            self.non_smooth_funs[0].eval(x),
            self.non_smooth_funs[1].eval(self.A(x)),
        ]
        return obj_nonsmooth + obj_smooth

    def _post(self):
        del self.z
        del self.u


class primal_dual(solver):
    r"""
    Parent class of all primal-dual algorithms.

    See generic attributes descriptions of the
    :class:`pyunlocbox.solvers.solver` base class.

    Parameters
    ----------
    L : function or ndarray, optional
        The transformation L that maps from the primal variable space to the
        dual variable space. Default is the identity, :math:`L(x)=x`. If `L` is
        an ``ndarray``, it will be converted to the operator form.
    Lt : function or ndarray, optional
        The adjoint operator. If `Lt` is an ``ndarray``, it will be converted
        to the operator form. If `L` is an ``ndarray``, default is the
        transpose of `L`. If `L` is a function, default is `L`,
        :math:`Lt(x)=L(x)`.
    d0: ndarray, optional
        Initialization of the dual variable.

    """

    def __init__(self, L=None, Lt=None, d0=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if L is None:
            self.L = lambda x: x
        else:
            if callable(L):
                self.L = L
            else:
                # Transform matrix form to operator form.
                self.L = lambda x: L.dot(x)

        if Lt is None:
            if L is None:
                self.Lt = lambda x: x
            elif callable(L):
                self.Lt = L
            else:
                self.Lt = lambda x: L.T.dot(x)
        else:
            if callable(Lt):
                self.Lt = Lt
            else:
                self.Lt = lambda x: Lt.dot(x)

        self.d0 = d0

    def _pre(self, functions, x0):
        # Dual variable.
        if self.d0 is None:
            # The copy is necessary in case `L = lambda x: x`.
            self.dual_sol = self.L(np.asarray(x0).copy())
        else:
            self.dual_sol = self.d0

    def _post(self):
        self.d0 = None
        del self.dual_sol

    def _objective(self, x):
        obj_smooth = [f.eval(x) for f in self.smooth_funs]
        obj_nonsmooth = [
            self.non_smooth_funs[0].eval(x),
            self.non_smooth_funs[1].eval(self.L(x)),
        ]
        return obj_nonsmooth + obj_smooth


class mlfbf(primal_dual):
    r"""
    Monotone+Lipschitz forward-backward-forward primal-dual algorithm.

    This algorithm solves convex optimization problems with objective of the
    form :math:`f(x) + g(Lx) + h(x)`, where :math:`f` and :math:`g` are proper,
    convex, lower-semicontinuous functions with easy-to-compute proximity
    operators, and :math:`h` has Lipschitz-continuous gradient with constant
    :math:`\beta`.

    See generic attributes descriptions of the
    :class:`pyunlocbox.solvers.primal_dual` base class.

    Notes
    -----
    The order of the functions matters: set :math:`f` first on the list,
    :math:`g` second, and :math:`h` third.

    This algorithm requires the first two functions to implement the
    :meth:`pyunlocbox.functions.func.prox` method, and the third function to
    implement the :meth:`pyunlocbox.functions.func.grad` method.

    The step-size should be in the interval :math:`\left] 0, \frac{1}{\beta +
    \|L\|_{2}}\right[`.

    See :cite:`komodakis2015primaldual`, Algorithm 6, for details.

    Examples
    --------
    >>> import numpy as np
    >>> from pyunlocbox import functions, solvers
    >>> y = np.array([294, 390, 361])
    >>> L = np.array([[5, 9, 3], [7, 8, 5], [4, 4, 9], [0, 1, 7]])
    >>> x0 = np.zeros(len(y))
    >>> f = functions.dummy()
    >>> f._prox = lambda x, T: np.maximum(np.zeros(len(x)), x)
    >>> g = functions.norm_l2(lambda_=0.5)
    >>> h = functions.norm_l2(y=y, lambda_=0.5)
    >>> max_step = 1/(1 + np.linalg.norm(L, 2))
    >>> solver = solvers.mlfbf(L=L, step=max_step/2.)
    >>> ret = solvers.solve([f, g, h], x0, solver, maxit=1000, rtol=0)
    Solution found after 1000 iterations:
        objective function f(sol) = 1.839060e+05
        stopping criterion: MAXIT
    >>> ret['sol']
    array([1., 1., 1.])

    """

    def _pre(self, functions, x0):
        super()._pre(functions, x0)

        if len(functions) != 3:
            raise ValueError("MLFBF requires 3 convex functions.")

        self.non_smooth_funs.append(functions[0])  # f
        self.non_smooth_funs.append(functions[1])  # g
        self.smooth_funs.append(functions[2])  # h

    def _algo(self):
        """
        y1 = x^k - λ ∇h(x^k) + L^T(z^k)
        y2 = z^k + λ L(x^k)

        p1 = prox_{λf} y1
        p2 = prox_{λg^*} y2

        q1 = p1 - λ ∇h(p1) + L^T(p2)
        q2 = p2 + λ L(p1)

        with x^k (z^k) the solution (dual-solution) at iteration k.
        """
        # Forward steps (in both primal and dual spaces)
        y1 = self.sol - self.step * (
            self.smooth_funs[0].grad(self.sol) + self.Lt(self.dual_sol)
        )
        y2 = self.dual_sol + self.step * self.L(self.sol)

        # Backward steps (in both primal and dual spaces)
        p1 = self.non_smooth_funs[0].prox(y1, self.step)
        p2 = _prox_star(self.non_smooth_funs[1], y2, self.step)

        # Forward steps (in both primal and dual spaces)
        q1 = p1 - self.step * (self.smooth_funs[0].grad(p1) + self.Lt(p2))
        q2 = p2 + self.step * self.L(p1)

        # Update solution (in both primal and dual spaces)
        self.sol[:] = self.sol - y1 + q1
        self.dual_sol[:] = self.dual_sol - y2 + q2


class projection_based(primal_dual):
    r"""
    Projection-based primal-dual algorithm.

    This algorithm solves convex optimization problems with objective of the
    form :math:`f(x) + g(Lx)`, where :math:`f` and :math:`g` are proper,
    convex, lower-semicontinuous functions with easy-to-compute proximity
    operators.

    See generic attributes descriptions of the
    :class:`pyunlocbox.solvers.primal_dual` base class.

    Parameters
    ----------
    lambda_ : float, optional
        The update term weight. It should be between 0 and 2. Default is 1.

    Notes
    -----
    The order of the functions matters: set :math:`f` first on the list, and
    :math:`g` second.

    This algorithm requires the two functions to implement the
    :meth:`pyunlocbox.functions.func.prox` method.

    The step-size should be in the interval :math:`\left] 0, \infty \right[`.

    See :cite:`komodakis2015primaldual`, Algorithm 7, for details.

    Examples
    --------
    >>> import numpy as np
    >>> from pyunlocbox import functions, solvers
    >>> y = np.array([294, 390, 361])
    >>> L = np.array([[5, 9, 3], [7, 8, 5], [4, 4, 9], [0, 1, 7]])
    >>> x0 = np.array([500, 1000, -400])
    >>> f = functions.norm_l1(y=y)
    >>> g = functions.norm_l1()
    >>> solver = solvers.projection_based(L=L, step=1.)
    >>> ret = solvers.solve([f, g], x0, solver, maxit=1000, rtol=None, xtol=.1)
    Solution found after 996 iterations:
        objective function f(sol) = 1.045000e+03
        stopping criterion: XTOL
    >>> ret['sol']
    array([0, 0, 0])

    """

    def __init__(self, lambda_=1.0, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lambda_ = lambda_

    def _pre(self, functions, x0):
        super()._pre(functions, x0)

        if self.lambda_ <= 0 or self.lambda_ > 2:
            raise ValueError("Lambda is bounded by 0 and 2.")

        if len(functions) != 2:
            raise ValueError("projection_based requires 2 convex functions.")

        self.non_smooth_funs.append(functions[0])  # f
        self.non_smooth_funs.append(functions[1])  # g

    def _algo(self):
        a = self.non_smooth_funs[0].prox(
            self.sol - self.step * self.Lt(self.dual_sol), self.step
        )
        ell = self.L(self.sol)
        b = self.non_smooth_funs[1].prox(ell + self.step * self.dual_sol, self.step)
        s = (self.sol - a) / self.step + self.Lt(ell - b) / self.step
        t = b - self.L(a)
        tau = np.sum(s**2) + np.sum(t**2)
        if tau == 0:
            self.sol[:] = a
            self.dual_sol[:] = self.dual_sol + (ell - b) / self.step
        else:
            theta = (
                self.lambda_
                * (
                    np.sum((self.sol - a) ** 2) / self.step
                    + np.sum((ell - b) ** 2) / self.step
                )
                / tau
            )
            self.sol[:] = self.sol - theta * s
            self.dual_sol[:] = self.dual_sol - theta * t


class chambolle_pock(primal_dual):
    r"""
    Chambolle-Pock Primal-Dual Proximal Splitting.

    This algorithm solves convex optimization problems with objective of the
    form :math:`G(x) + F(Lx)`, where :math:`F` and :math:`G` are proper,
    convex, lower-semicontinuous functions with easy-to-compute proximity
    operators, and :math:`L` is a linear operator.

    See generic attributes descriptions of the
    :class:`pyunlocbox.solvers.primal_dual` base class.

    Notes
    -----
    The order of the functions matters: set :math:`G` first on the list and
    :math:`F` second.

    This algorithm requires the two functions to implement the
    :meth:`pyunlocbox.functions.func.prox` method.

    The step-size should be in the interval
    :math:`\left] 0, \frac{1}{\beta + \|L\|_{2}}\right[`.

    See [1]_ for details.

    References
    ----------
    .. [1] Antonin Chambolle and Thomas Pock, "A First-order primal-dual
       algorithm for convex problems with application to imaging," Journal of
       Mathematical Imaging and Vision, Volume 40, Number 1 (2011), 120-145.

    Examples
    --------
    >>> import numpy as np
    >>> from pyunlocbox import functions, solvers
    >>> y = np.array([4, -9, -13, -4])
    >>> L = np.array([[5, 9, 3], [7, 8, 5], [4, 4, 9], [0, 1, 7]])
    >>> max_step = 1 / (1 + np.linalg.norm(L, 2))
    >>> x0 = np.zeros(3)
    >>> f = functions.norm_l1(y=y)
    >>> g = functions.norm_l1()
    >>> solver = solvers.chambolle_pock(L=L, sigma=max_step/2,
    ...                                 theta=max_step/2, tau=max_step/2)
    >>> ret = solvers.solve([g, f], x0, solver, maxit=1000, rtol=None, xtol=None)
    Solution found after 1000 iterations:
        objective function f(sol) = 9.068105e+00
        stopping criterion: MAXIT
    >>> ret['sol']
    array([-3.60833292,  2.72982626, -1.02656965])

    """

    def __init__(self, sigma=1.0, tau=1.0, theta=1.0, accel=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.sigma = sigma
        self.tau = tau
        self.theta = theta
        # self.accel = acceleration.dummy() if accel is None else accel

    def _pre(self, functions, x0):
        super()._pre(functions, x0)

        if self.tau <= 0 or self.tau > 2:
            raise ValueError("tau is bounded by 0 and 2.")
        if self.sigma <= 0 or self.sigma > 2:
            raise ValueError("sigma is bounded by 0 and 2.")
        if self.theta <= 0 or self.theta > 2:
            raise ValueError("theta is bounded by 0 and 2.")

        if len(functions) != 2:
            raise ValueError("Chambolle-Pock requires 2 functions.")

        self.non_smooth_funs.append(functions[0])  # F
        self.non_smooth_funs.append(functions[1])  # G

        # Initializations
        self.f = np.array(x0, copy=True)
        self.g = np.array(self.L(x0), copy=True)

    def _algo(self):
        """
        g_{k+1} = prox_{\\sigma F^∗} (g_k+ \\sigma*L(\tilde{f}_k))
        f_{k+1} = prox_{\tau G} (f_k−\tau L^∗(g_{k+1}) )
        \tilde{f}_{k+1} = f{k+1}+ \theta*(f_{k+1} − f_k)
        """
        # Backward steps
        self.g = _prox_star(
            self.non_smooth_funs[1], self.g + self.sigma * self.L(self.sol), self.sigma
        )
        self.fp1 = self.non_smooth_funs[0].prox(
            self.f - self.tau * self.Lt(self.g), self.tau
        )

        # Update solution
        self.sol[:] = self.fp1 + self.theta * (self.fp1 - self.f)

        # update
        self.f = np.copy(self.fp1)

    def _post(self):
        del self.g, self.fp1, self.f



#!/usr/bin/env python

"""
Test suite for the acceleration module of the pyunlocbox package.

"""

import numpy as np
import numpy.testing as nptest
import pytest

from pyunlocbox import acceleration, functions, solvers


class TestAcceleration:

    def test_accel(self):
        """
        Test base acceleration scheme class

        """
        funs = [functions.dummy(), functions.dummy()]
        x0 = np.zeros((4,))
        a = acceleration.accel()
        s = solvers.forward_backward()
        o = [[1.0, 2.0], [0.0, 1.0]]
        n = 2

        with pytest.raises(NotImplementedError):
            a.pre(funs, x0)
        with pytest.raises(NotImplementedError):
            a.update_step(s, o, n)
        with pytest.raises(NotImplementedError):
            a.update_sol(s, o, n)
        with pytest.raises(NotImplementedError):
            a.post()

    def test_backtracking(self):
        """
        Test forward-backward splitting solver with backtracking, solving
        problems with L1-norm, L2-norm, and dummy functions.

        """
        # Test constructor sanity
        a = acceleration.backtracking()
        with pytest.raises(ValueError):
            a.__init__(2.0)
        with pytest.raises(ValueError):
            a.__init__(-2.0)

        y = [4.0, 5.0, 6.0, 7.0]
        accel = acceleration.backtracking()
        step = 10  # Make sure backtracking is called
        solver = solvers.forward_backward(accel=accel, step=step)
        param = {"solver": solver, "atol": 1e-32, "verbosity": "NONE"}

        # L2-norm prox and dummy gradient.
        f1 = functions.norm_l2(y=y)
        f2 = functions.dummy()
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret["sol"], y)
        assert ret["crit"] == "ATOL"
        assert ret["niter"] == 13

        # L1-norm prox and L2-norm gradient.
        f1 = functions.norm_l1(y=y, lambda_=1.0)
        f2 = functions.norm_l2(y=y, lambda_=0.8)
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret["sol"], y)
        assert ret["crit"] == "ATOL"
        assert ret["niter"] <= 4  # win64 takes one iteration

    def test_forward_backward_fista(self):
        """
        Test forward-backward splitting solver with fista acceleration,
        solving problems with L1-norm, L2-norm, and dummy functions.

        """
        y = [4.0, 5.0, 6.0, 7.0]
        solver = solvers.forward_backward(accel=acceleration.fista())
        param = {"solver": solver, "rtol": 1e-6, "verbosity": "NONE"}

        # L2-norm prox and dummy gradient.
        f1 = functions.norm_l2(y=y)
        f2 = functions.dummy()
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret["sol"], y)
        assert ret["crit"] == "RTOL"
        assert ret["niter"] == 60

        # Dummy prox and L2-norm gradient.
        f1 = functions.dummy()
        f2 = functions.norm_l2(y=y, lambda_=0.6)
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret["sol"], y)
        assert ret["crit"] == "RTOL"
        assert ret["niter"] == 84

        # L2-norm prox and L2-norm gradient.
        f1 = functions.norm_l2(y=y)
        f2 = functions.norm_l2(y=y)
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret["sol"], y, rtol=1e-2)
        assert ret["crit"] == "MAXIT"
        assert ret["niter"] == 200

        # L1-norm prox and dummy gradient.
        f1 = functions.norm_l1(y=y)
        f2 = functions.dummy()
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret["sol"], y)
        assert ret["crit"] == "RTOL"
        assert ret["niter"] == 6

        # Dummy prox and L1-norm gradient. As L1-norm possesses no gradient,
        # the algorithm exchanges the functions : exact same solution.
        f1 = functions.dummy()
        f2 = functions.norm_l1(y=y)
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret["sol"], y)
        assert ret["crit"] == "RTOL"
        assert ret["niter"] == 6

        # L1-norm prox and L1-norm gradient. L1-norm possesses no gradient.
        f1 = functions.norm_l1(y=y)
        f2 = functions.norm_l1(y=y)
        with pytest.raises(ValueError):
            solvers.solve([f1, f2], np.zeros(len(y)), **param)

        # L1-norm prox and L2-norm gradient.
        f1 = functions.norm_l1(y=y, lambda_=1.0)
        f2 = functions.norm_l2(y=y, lambda_=0.8)
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret["sol"], y)
        assert ret["crit"] == "RTOL"
        assert ret["niter"] == 10

    def test_forward_backward_fista_backtracking(self):
        """
        Test forward-backward splitting solver with fista acceleration and
        backtracking, solving problems with L1-norm, L2-norm, and dummy
        functions.

        """
        y = [4.0, 5.0, 6.0, 7.0]
        accel = acceleration.fista_backtracking()
        solver = solvers.forward_backward(accel=accel)
        param = {"solver": solver, "rtol": 1e-6, "verbosity": "NONE"}

        # L2-norm prox and dummy gradient.
        f1 = functions.norm_l2(y=y)
        f2 = functions.dummy()
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret["sol"], y)
        assert ret["crit"] == "RTOL"
        assert ret["niter"] == 60

        # L1-norm prox and L2-norm gradient.
        f1 = functions.norm_l1(y=y, lambda_=1.0)
        f2 = functions.norm_l2(y=y, lambda_=0.8)
        ret = solvers.solve([f1, f2], np.zeros(len(y)), **param)
        nptest.assert_allclose(ret["sol"], y)
        assert ret["crit"] == "RTOL"
        assert ret["niter"] == 3

    def test_regularized_nonlinear(self):
        """
        Test gradient descent solver with regularized non-linear acceleration,
        solving problems with L2-norm functions.

        """
        dim = 25
        np.random.seed(0)
        x0 = np.random.rand(dim)
        xstar = np.random.rand(dim)
        x0 = xstar + 5.0 * (x0 - xstar) / np.linalg.norm(x0 - xstar)

        A = np.random.rand(dim, dim)
        step = 1 / np.linalg.norm(np.dot(A.T, A))

        accel = acceleration.regularized_nonlinear(k=5)
        solver = solvers.gradient_descent(step=step, accel=accel)
        param = {"solver": solver, "rtol": 0, "maxit": 200, "verbosity": "NONE"}

        # L2-norm prox and dummy gradient.
        f1 = functions.norm_l2(lambda_=0.5, A=A, y=np.dot(A, xstar))
        f2 = functions.dummy()
        ret = solvers.solve([f1, f2], x0, **param)
        pctdiff = 100 * np.sum((xstar - ret["sol"]) ** 2) / np.sum(xstar**2)
        nptest.assert_array_less(pctdiff, 1.91)

        # Sanity checks
        accel = acceleration.regularized_nonlinear()
        with pytest.raises(ValueError):
            accel.__init__(10, ["not", "good"])
        with pytest.raises(ValueError):
            accel.__init__(10, "nope")

    def test_acceleration_comparison(self):
        """
        Test that all solvers return the same and correct solution.

        """

        # Convex functions.
        y = [1, 0, 0.1, 8, -6.5, 0.2, 0.004, 0.01]
        sol = [0.75, 0, 0, 7.75, -6.25, 0, 0, 0]
        w1, w2 = 0.8, 0.4
        f1 = functions.norm_l2(y=y, lambda_=w1 / 2.0)  # Smooth.
        f2 = functions.norm_l1(lambda_=w2 / 2.0)  # Non-smooth.

        # Solvers.
        L = w1  # Lipschitz continuous gradient.
        step = 1.0 / L
        slvs = []
        slvs.append(solvers.forward_backward(accel=acceleration.dummy(), step=step))
        slvs.append(solvers.forward_backward(accel=acceleration.fista(), step=step))
        slvs.append(
            solvers.forward_backward(
                accel=acceleration.fista_backtracking(eta=0.999), step=step
            )
        )

        # Compare solutions.
        params = {"rtol": 1e-14, "verbosity": "NONE", "maxit": 1e4}
        niters = [2, 2, 6]
        for solver, niter in zip(slvs, niters):
            x0 = np.zeros(len(y))
            ret = solvers.solve([f1, f2], x0, solver, **params)
            nptest.assert_allclose(ret["sol"], sol)
            assert ret["niter"] == niter



"""
Test suite for the docstrings of the pyunlocbox package.

"""

import doctest
import os

import pytest


def gen_recursive_file(root, ext):
    """Generate file paths recursively."""
    for root, _, filenames in os.walk(root):
        for name in filenames:
            if name.lower().endswith(ext):
                yield os.path.join(root, name)


def test_docstrings_python():
    """Test docstrings in Python files."""
    files = list(gen_recursive_file("pyunlocbox", ".py"))
    for file in files:
        try:
            doctest.testfile(file, module_relative=False, verbose=False)
        except doctest.DocTestFailure as e:
            pytest.fail(f"Doctest failed in {file}: {e}")
        except Exception:
            # Some files might not have doctests or might have import issues
            # We'll skip those for now
            pass


def test_docstrings_rst():
    """Test docstrings in RST files."""
    files = list(gen_recursive_file(".", ".rst"))
    # Filter out files that are likely to have import issues
    files = [
        f
        for f in files
        if not any(exclude in f for exclude in [".venv", "build", "_build"])
    ]

    for file in files:
        try:
            doctest.testfile(file, module_relative=False, verbose=False)
        except doctest.DocTestFailure as e:
            pytest.fail(f"Doctest failed in {file}: {e}")
        except Exception:
            # Some files might not have doctests or might have import issues
            # We'll skip those for now
            pass



#!/usr/bin/env python

"""
Test suite for the functions module of the pyunlocbox package.

"""

import inspect

import numpy as np
import numpy.testing as nptest
import pytest

from pyunlocbox import functions


class TestFunctions:

    def test_func(self):
        """
        Test the func base class.
        First test that all the methods raise a NotImplemented exception.
        Then assign valid methods and test return values.

        """
        x = 10
        T = 1
        f = functions.func()
        with pytest.raises(NotImplementedError):
            f.eval(x)
        with pytest.raises(NotImplementedError):
            f.prox(x, T)
        with pytest.raises(NotImplementedError):
            f.grad(x)
        assert len(f.cap(x)) == 0
        # Set up but never used:
        # f.eval = lambda x: x**2 - 5
        # f.grad = lambda x: 2 * x
        # f.prox = lambda x, T: x + T

        def assert_equivalent(param1, param2):
            x = [[7, 8, 9], [10, 324, -45], [-7, -0.2, 5]]
            for name, func in inspect.getmembers(functions, inspect.isclass):
                if name not in ["func", "norm", "proj"]:
                    f1 = func(**param1)
                    f2 = func(**param2)
                    assert f1.eval(x) == f2.eval(x)
                    nptest.assert_array_equal(f1.prox(x, 3), f2.prox(x, 3))
                    if "GRAD" in f1.cap(x):
                        nptest.assert_array_equal(f1.grad(x), f2.grad(x))

        # Default parameters. Callable or matrices.
        assert_equivalent({"y": 3.2}, {"y": lambda: 3.2})
        assert_equivalent({"A": None}, {"A": np.identity(3)})
        A = np.array([[-4, 2, 5], [1, 3, -7], [2, -1, 0]])
        pinvA = np.linalg.pinv(A)  # For proj_linalg.
        assert_equivalent({"A": A}, {"A": A, "At": A.T})
        assert_equivalent({"A": lambda x: A.dot(x), "pinvA": pinvA}, {"A": A, "At": A})

    def test_dummy(self):
        """
        Test the dummy derived class.
        All the methods should return 0.

        """
        f = functions.dummy()
        assert f.eval(34) == 0
        nptest.assert_array_equal(f.grad(34), [0])
        nptest.assert_array_equal(f.prox(34, 1), [34])
        x = [34, 2, 1.0, -10.2]
        assert f.eval(x) == 0
        nptest.assert_array_equal(f.grad(x), np.zeros(len(x)))
        nptest.assert_array_equal(f.prox(x, 1), x)

    def test_norm_l2(self):
        """
        Test the norm_l2 derived class.
        We test the three methods : eval, grad and prox.
        First with default class properties, then custom ones.

        """
        f = functions.norm_l2(lambda_=3)
        assert f.eval([10, 0]) == 300
        assert f.eval(np.array([-10, 0])) == 300
        nptest.assert_allclose(f.grad([10, 0]), [60, 0])
        nptest.assert_allclose(f.grad([-10, 0]), [-60, 0])
        assert f.eval([3, 4]) == 3 * 5**2
        assert f.eval(np.array([-3, 4])) == 3 * 5**2
        nptest.assert_allclose(f.grad([3, 4]), [18, 24])
        nptest.assert_allclose(f.grad([3, -4]), [18, -24])
        assert f.prox(0, 1) == 0
        assert f.prox(7, 1.0 / 6) == 3.5
        f = functions.norm_l2(lambda_=4)
        nptest.assert_allclose(f.prox([7, -22], 0.125), [3.5, -11])

        f = functions.norm_l2(
            lambda_=1, A=lambda x: 2 * x, At=lambda x: x / 2, y=[8, 12]
        )
        assert f.eval([4, 6]) == 0
        assert f.eval([5, -2]) == 256 + 4
        nptest.assert_allclose(f.grad([4, 6]), 0)
        # nptest.assert_allclose(f.grad([5, -2]), [8, -64])
        nptest.assert_allclose(f.prox([4, 6], 1), [4, 6])

        f = functions.norm_l2(
            lambda_=2,
            y=np.fft.fft([2, 4]) / np.sqrt(2),
            A=lambda x: np.fft.fft(x) / np.sqrt(x.size),
            At=lambda x: np.fft.ifft(x) * np.sqrt(x.size),
        )
        # assert f.eval(np.fft.ifft([2, 4]*np.sqrt(2)), 0)
        # assert f.eval([3, 5], 2*np.sqrt(25+81))
        nptest.assert_allclose(f.grad([2, 4]), 0)
        # nptest.assert_allclose(f.grad([3, 5]), [4*np.sqrt(5), 4*3])
        nptest.assert_allclose(f.prox([2, 4], 1), [2, 4])
        nptest.assert_allclose(f.prox([3, 5], 1), [2.2, 4.2])
        nptest.assert_allclose(f.prox([2.2, 4.2], 1), [2.04, 4.04])
        nptest.assert_allclose(f.prox([2.04, 4.04], 1), [2.008, 4.008])

        # Test prox for non-tight matrices A
        L = np.array([[8, 1, 10], [1, 9, 1], [3, 7, 5], [1, 4, 4]])
        f = functions.norm_l2(
            A=L, tight=False, y=np.array([1, 2, 3, 4]), w=np.array([1, 1, 0.5, 0.75])
        )
        nptest.assert_allclose(f.eval([1, 1, 1]), 455.0625)
        nptest.assert_allclose(
            f.grad([1, 1, 1]), [329.625, 262.500, 430.500], rtol=1e-3
        )
        nptest.assert_allclose(f.prox([1, 1, 1], 1), [-0.887, 0.252, 0.798], rtol=1e-3)
        nptest.assert_allclose(f.prox([6, 7, 3], 1), [-0.345, 0.298, 0.388], rtol=1e-3)
        nptest.assert_allclose(
            f.prox([10, 0, -5], 1), [1.103, 0.319, -0.732], rtol=1e-3
        )

        # Cannot have non-scalar weights with tight operator.
        with pytest.raises(ValueError):
            functions.norm_l2(
                A=L, tight=True, y=np.array([1, 2, 3, 4]), w=np.array([1, 1, 0.5, 0.75])
            )
        functions.norm_l2(A=L, tight=True, y=[1, 2, 3, 4], w=np.array([2]))
        functions.norm_l2(A=L, tight=True, y=[1, 2, 3, 4], w=2)

    def test_soft_thresholding(self):
        """
        Test the soft thresholding helper function.

        """
        x = np.arange(-4, 5, 1)
        # Test integer division for complex method.
        Ts = [2]
        y_gold = [[-2, -1, 0, 0, 0, 0, 0, 1, 2]]
        # Test division by 0 for complex method.
        Ts.append([0.4, 0.3, 0.2, 0.1, 0, 0.1, 0.2, 0.3, 0.4])
        y_gold.append([-3.6, -2.7, -1.8, -0.9, 0, 0.9, 1.8, 2.7, 3.6])
        for k, T in enumerate(Ts):
            for cmplx in [False, True]:
                y_test = functions._soft_threshold(x, T, cmplx)
                nptest.assert_array_equal(y_test, y_gold[k])

    def test_norm_l1(self):
        """
        Test the norm_l1 derived class.
        We test the two methods : eval and prox.
        First with default class properties, then custom ones.

        """
        f = functions.norm_l1(lambda_=3)
        assert f.eval([10, 0]) == 30
        assert f.eval(np.array([-10, 0])) == 30
        assert f.eval([-3, 4]) == 21
        nptest.assert_array_equal(
            f.prox(np.array([[1, -4], [5, -2]]), 1), [[0, -1], [2, 0]]
        )

        f = functions.norm_l1(tight=False)
        with pytest.raises(NotImplementedError):
            f.prox([1, 2], 1)

        # Let us test the weights
        f2 = functions.norm_l1(lambda_=4, w=2)
        assert f2.eval([10, 1]) == 88
        assert f2.eval([5]) == 40
        assert f2.eval([0, -2, 1]) == 24
        assert f2.eval(np.array([-10, 0])) == 80
        assert f2.eval([-3, 4]) == 56
        nptest.assert_array_equal(
            f2.prox(np.array([[7, -9], [10, -2]]), 1), [[0, -1], [2, 0]]
        )

        # Let us multiple weights
        f3 = functions.norm_l1(lambda_=4, w=[1, 2])
        assert f3.eval([10, 1]) == 48
        nptest.assert_array_equal(
            f3.prox(np.array([[7, -9], [10, -2]]), 1), [[3, -1], [6, 0]]
        )

        # Let us test y
        f3 = functions.norm_l1(lambda_=2, y=[1, 2])
        assert f3.eval([10, 1]) == 20
        nptest.assert_array_equal(
            f3.prox(np.array([[2, 5], [-1, -2]]), 1), [[1, 3], [1, 0]]
        )

        # Forward operators (square and non-square).
        A = np.eye(3)
        f = functions.norm_l1(A=A)
        assert f.eval([1, 2, 4]) == 7
        nptest.assert_allclose(f.prox([1, 2, 4], 1), [0, 1, 3])
        A = np.concatenate([A] * 2, 0)
        f = functions.norm_l1(A=A)
        assert f.eval([1, 2, 4]) == 14
        nptest.assert_allclose(f.prox([1, 2, 4], 1), [-1, 0, 2])

    def test_norm_nuclear(self):
        """
        Test the norm_nuclear derived class.
        We test the two methods : eval and prox.
        First with default class properties, then custom ones.

        """

        # TODO: this method should be tested more extensively.
        f = functions.norm_nuclear(lambda_=3)
        assert f.eval(np.diag([10, 0])) == 30
        assert f.eval(np.diag(np.array([-10, 0]))) == 30
        assert f.eval([[-3]]) == 9
        nptest.assert_allclose(
            f.prox(np.array([[1, 1], [1, 1]]), 1.0 / 3), [[0.5, 0.5], [0.5, 0.5]]
        )

    def test_norm_tv(self):
        """
        Test the norm_tv derived class.
        We test the grad, div, eval and prox.

        """
        # Test Matrices initialization
        # test for a 1dim matrice (testing with a 5)
        # mat1d = np.arange(5) + 1
        # test for a 2dim matrice (testing with a 2x4)
        mat2d = np.array([[2, 3, 0, 1], [22, 1, 4, 5]])
        # test for a 3 dim matrice (testing with a 2x3x2)
        mat3d = np.arange(1, stop=13).reshape(2, 2, 3).transpose((1, 2, 0))
        # test for a 4dim matrice (2x3x2x2)
        # mat4d = np.arange(1, stop=25).reshape(2, 2, 2, 3)
        # mat4d = mat4d.transpose((2, 3, 1, 0))
        # test for a 5dim matrice (2x2x3x2x2)
        # mat5d = np.arange(1, stop=49).reshape(2, 2, 3, 2, 2)
        # mat5d = mat5d.transpose((3, 4, 2, 1, 0))

        # Test for evals
        def test_eval():

            # test with 2d matrices
            # test without weight
            f = functions.norm_tv(dim=1)
            xeval = 30
            nptest.assert_array_equal(xeval, f.eval(mat2d))
            f = functions.norm_tv(dim=2)
            xeval = np.array([56.753641295582440])
            nptest.assert_array_equal(xeval, f.eval(mat2d))

            # test with weights
            f = functions.norm_tv(dim=1, wx=3)
            xeval = np.sum(np.array([60, 6, 12, 12]))
            nptest.assert_array_equal(xeval, f.eval(mat2d))
            f = functions.norm_tv(dim=2, wx=0.5, wy=2)
            xeval = np.array([71.1092])
            nptest.assert_array_equal(xeval, np.around(f.eval(mat2d), decimals=4))

            # test with 3d matrices (2x3x2)
            # test without weight
            f = functions.norm_tv(dim=2)
            sol = np.sum(np.array([11.324555320336760, 11.324555320336760]))
            nptest.assert_array_equal(sol, f.eval(mat3d))
            f = functions.norm_tv(dim=3)
            xeval = np.array(49.762944279683104)
            nptest.assert_array_equal(xeval, f.eval(mat3d))

            # test with weights
            f = functions.norm_tv(dim=2, wx=2, wy=3)
            sol = np.sum(np.array([25.4164, 25.4164]))
            nptest.assert_array_equal(sol, np.around(f.eval(mat3d), decimals=4))

            f = functions.norm_tv(dim=3, wx=2, wy=3, wz=0.5)
            xeval = np.array([58.3068])
            nptest.assert_array_equal(xeval, np.around(f.eval(mat3d), decimals=4))

        # Test for prox
        def test_prox():

            # Test with 2d matrices
            # Test without weights
            f = functions.norm_tv(tol=10e-4, dim=1)
            gamma = 30
            sol = np.array(
                [
                    [
                        12.003459453582762,
                        1.999654054641723,
                        2.000691890716554,
                        3.000691890716554,
                    ],
                    [
                        11.996540546417238,
                        2.000345945358277,
                        1.999308109283446,
                        2.999308109283446,
                    ],
                ]
            )
            nptest.assert_array_equal(
                np.around(sol, decimals=5),
                np.around((f.prox(mat2d, gamma)), decimals=5),
            )

            f = functions.norm_tv(tol=10e-4, dim=2)
            gamma = 1.5
            x2d = np.array([[2, 3, 0, 1], [22, 1, 4, 5], [2, 10, 7, 8]])
            sol = np.array(
                [
                    [3.44427, 2.87332, 2.51662, 2.45336],
                    [18.38207, 3.10251, 4.0028, 4.64074],
                    [4.50809, 6.44118, 6.38421, 6.25082],
                ]
            )
            nptest.assert_array_equal(
                np.around(sol, decimals=5), np.around((f.prox(x2d, gamma)), decimals=5)
            )

            # Test with weights

            # Test with 3d matrices
            # Test without weights
            f = functions.norm_tv(tol=10e-4, dim=2)
            gamma = 42
            sol = np.array(
                [
                    [[3.50087, 9.50087], [3.50000, 9.50000], [3.49913, 9.49913]],
                    [[3.50087, 9.50087], [3.50000, 9.50000], [3.49913, 9.49913]],
                ]
            )
            nptest.assert_array_equal(sol, np.around(f.prox(mat3d, gamma), decimals=5))

            f = functions.norm_tv(tol=10e-4, dim=3)
            gamma = 18
            sol = np.array(
                [
                    [[6.5, 6.5], [6.5, 6.5], [6.5, 6.5]],
                    [[6.5, 6.5], [6.5, 6.5], [6.5, 6.5]],
                ]
            )
            nptest.assert_array_equal(sol, np.around(f.prox(mat3d, gamma), decimals=1))
            # Test with weights
            f = functions.norm_tv(tol=10e-10, dim=2, wx=5, wy=10)
            gamma = 3
            x3d = np.array(
                [
                    [[1, 10, 19], [2, 11, 20], [3, 12, 21]],
                    [[4, 13, 22], [5, 14, 23], [6, 15, 24]],
                    [[7, 16, 25], [8, 17, 26], [9, 18, 27]],
                ]
            )
            sol = np.array(
                [
                    [[5, 14, 23], [5, 14, 23], [5, 14, 23]],
                    [[5, 14, 23], [5, 14, 23], [5, 14, 23]],
                    [[5, 14, 23], [5, 14, 23], [5, 14, 23]],
                ]
            )
            nptest.assert_array_equal(sol, np.around(f.prox(x3d, gamma)))

            # Test with 4d matrices
            # Test without weights
            f = functions.norm_tv(tol=10e-4, dim=3)
            gamma = 10
            x4d = np.array(
                [
                    [
                        [[1, 28, 55], [10, 37, 64], [19, 46, 73]],
                        [[2, 29, 56], [11, 38, 65], [20, 47, 74]],
                        [[3, 30, 57], [12, 39, 66], [21, 48, 75]],
                    ],
                    [
                        [[4, 31, 58], [13, 40, 67], [22, 49, 76]],
                        [[5, 32, 59], [14, 41, 68], [23, 50, 77]],
                        [[6, 33, 60], [15, 42, 69], [24, 51, 78]],
                    ],
                    [
                        [[7, 34, 61], [16, 43, 70], [25, 52, 79]],
                        [[8, 35, 62], [17, 44, 71], [26, 53, 80]],
                        [[9, 36, 63], [18, 45, 72], [27, 54, 81]],
                    ],
                ]
            )
            sol = np.array(
                [
                    [
                        [[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                        [[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                        [[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                    ],
                    [
                        [[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                        [[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                        [[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                    ],
                    [
                        [[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                        [[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                        [[14, 41, 68], [14, 41, 68], [14, 41, 68]],
                    ],
                ]
            )
            sol = np.around(sol)
            nptest.assert_array_equal(sol, np.around(f.prox(x4d, gamma)))

            f = functions.norm_tv(tol=10e-4, dim=4)
            gamma = 15
            sol = np.array(
                [
                    [
                        [[22, 34, 54], [26, 40, 54], [31, 44, 53]],
                        [[23, 35, 54], [27, 40, 54], [32, 44, 53]],
                        [[23, 35, 54], [27, 41, 54], [32, 45, 53]],
                    ],
                    [
                        [[24, 36, 54], [28, 41, 54], [32, 45, 53]],
                        [[24, 36, 54], [28, 42, 53], [33, 45, 53]],
                        [[24, 37, 54], [29, 42, 53], [33, 46, 53]],
                    ],
                    [
                        [[25, 38, 54], [29, 43, 53], [34, 46, 53]],
                        [[25, 38, 54], [30, 43, 53], [34, 46, 53]],
                        [[26, 39, 54], [30, 43, 53], [35, 47, 53]],
                    ],
                ]
            )
            sol = np.around(sol)
            nptest.assert_array_equal(sol, np.around(f.prox(x4d, gamma)))

        # Test with weights
        test_prox()
        test_eval()

    def test_proj_b2(self):
        """
        Test the projection on the L2-ball.
        ISTA and FISTA algorithms for tight and non-tight frames.

        """
        tol = 1e-7

        # Tight frame, radius 0 --> x == y.
        x = np.random.uniform(size=7) + 10
        y = np.random.uniform(size=7) - 10
        f = functions.proj_b2(y=y, epsilon=tol)
        nptest.assert_allclose(f.prox(x, 0), y, atol=tol)

        # Tight frame, random radius --> ||x-y||_2 = radius.
        radius = np.random.uniform()
        f = functions.proj_b2(y=y, epsilon=radius)
        nptest.assert_almost_equal(np.linalg.norm(f.prox(x, 0) - y), radius)

        # Always evaluate to zero.
        assert f.eval(x) == 0

        # Non-tight frame : compare FISTA and ISTA results.
        nx = 30
        ny = 15
        x = np.random.standard_normal(nx)
        y = np.random.standard_normal(ny)
        A = np.random.standard_normal((ny, nx))
        nu = np.linalg.norm(A, ord=2) ** 2
        f = functions.proj_b2(
            y=y, A=A, nu=nu, tight=False, method="FISTA", epsilon=5, tol=tol / 10
        )
        sol_fista = f.prox(x, 0)
        f.method = "ISTA"
        sol_ista = f.prox(x, 0)
        nptest.assert_allclose(sol_fista, sol_ista, rtol=1e-3)

        f.method = "NOT_A_VALID_METHOD"
        with pytest.raises(ValueError):
            f.prox(x, 0)

    def test_proj_lineq(self):
        """
        Test the projection on Ax = y

        """
        x = np.zeros([10])
        A = np.ones([1, 10])
        y = np.array([10])
        f = functions.proj_lineq(A=A, y=y)
        sol = f.prox(x, 0)
        np.testing.assert_allclose(sol, np.ones([10]))
        np.testing.assert_allclose(A.dot(sol), y)

        f = functions.proj_lineq(A=A)
        sol = f.prox(x, 0)
        np.testing.assert_allclose(sol, np.zeros([10]))

        for i in range(1, 15):
            x = np.random.randn(10)
            y = np.random.randn(i)
            A = np.random.randn(i, 10)
            pinvA = np.linalg.pinv(A)
            f1 = functions.proj_lineq(A=A, y=y)
            f2 = functions.proj_lineq(A=lambda x: A.dot(x), pinvA=pinvA, y=y)
            f3 = functions.proj_lineq(A=A, pinvA=lambda x: pinvA.dot(x), y=y)
            f4 = functions.proj_lineq(A=A, pinvA=pinvA, y=y)
            sol1 = f1.prox(x, 0)
            sol2 = f2.prox(x, 0)
            sol3 = f3.prox(x, 0)
            sol4 = f4.prox(x, 0)
            np.testing.assert_allclose(sol1, sol2)
            np.testing.assert_allclose(sol1, sol3)
            np.testing.assert_allclose(sol1, sol4)
            if i <= x.size:
                np.testing.assert_allclose(A.dot(sol1), y)
            if i >= x.size:
                np.testing.assert_allclose(sol1, pinvA.dot(y))

        with pytest.raises(ValueError):
            functions.proj_lineq(A=lambda x: x)

    def test_proj_positive(self):
        """
        Test the projection on the positive octant.

        """
        fpos = functions.proj_positive()
        x = np.random.randn(10, 12)
        res = fpos.prox(x, T=1)
        nptest.assert_equal(res >= 0, True)  # All values are positive.
        nptest.assert_equal(res[x < 0], 0)  # Negative values are set to zero.
        nptest.assert_equal(res[x > 0], x[x > 0])  # Positives are unchanged.
        assert fpos.eval(x) == 0

    def test_proj_spsd(self):
        """
        Test the projection on symmetric positive semi-definite matrices.

        """
        f_spds = functions.proj_spsd()
        A = np.random.randn(10, 10)
        A = A + A.T
        eig1 = np.sort(np.real(np.linalg.eig(A)[0]))
        res = f_spds.prox(A, T=1)
        eig2 = np.sort(np.real(np.linalg.eig(res)[0]))
        # All eigenvalues are positive.
        assert (eig2 > -1e-13).all()

        # Positive value are unchanged.
        np.testing.assert_allclose(eig2[eig1 > 0], eig1[eig1 > 0])

        # The symmetrization works.
        A = np.random.rand(10, 10) + 10 * np.eye(10)
        res = f_spds.prox(A, T=1)
        np.testing.assert_allclose(res, (A + A.T) / 2)

        assert f_spds.eval(A) == 0

    def test_structured_sparsity(self):
        """
        Test the structured sparsity function.

        """
        # test init
        # test parsing of lambda
        with pytest.raises(ValueError):
            functions.structured_sparsity(-1, [[]], [0.0])
        # test parsing of groups
        with pytest.raises(TypeError):
            functions.structured_sparsity(1.0, 1, [0.0])
        # test parsing of weights
        with pytest.raises(ValueError):
            functions.structured_sparsity(1.0, [[1, 2], [3, 4]], [10.0])

        # test call of eval and prox
        x = np.array([0.01, 0.5, 3, 4])
        groups = [[0, 1], [2, 3]]
        weights = np.array([10, 0.2])
        f = functions.structured_sparsity(1, groups, weights)
        # test eval
        result = f.eval(x)
        gt = weights[0] * np.linalg.norm(x[groups[0]], 2) + weights[1] * np.linalg.norm(
            x[groups[1]], 2
        )
        assert result == pytest.approx(gt)
        # test prox
        gt = x.copy()
        # the first group has norm lower than the corresponding weight
        gt[groups[0]] = 0
        # the second group has norm higher than the corresponding weight
        gt[groups[1]] -= x[groups[1]] * weights[1] / np.linalg.norm(x[groups[1]])
        prox = f.prox(x, 1)
        np.testing.assert_almost_equal(prox, gt)

    def test_capabilities(self):
        """
        Test that a function implements the necessary methods. A function must
        always be evaluable (by implementing _eval) and must have a gradient or
        proximal operator (by implementing at least one of _prox or _grad).

        """
        for name, func in inspect.getmembers(functions, inspect.isclass):
            if name in ["func", "proj", "norm"]:  # exclude abstract classes
                continue
            cap = func().cap(np.diag([10, 42]))
            assert "EVAL" in cap
            assert "GRAD" in cap or "PROX" in cap

    def test_independent_problems(self):
        """
        Test that multiple independent problems can be solved in parallel.

        """

        # Parameters.
        N = 3  # independent problems.
        n = 25  # dimensions.

        # Generate some data.
        X = 7 - 10 * np.random.uniform(size=(n, N))
        step = 10 * np.random.uniform()

        # Test all available functions.
        for name, func in inspect.getmembers(functions, inspect.isclass):
            # TODO: use subTest once python 2.7 is dropped
            # with self.subTest(i=func[0]):

            # Instantiate the class.
            if name == "norm_tv":
                # Each column is one-dimensional.
                f = func(dim=1, maxit=20, tol=0)
            elif name in ["norm_nuclear", "proj_spsd"]:
                # TODO: make this test two dimensional for the norm nuclear
                # and the spsd projection?
                continue
            else:
                f = func()

            cap = f.cap(X)

            # The combined objective function of the N problems is the sum of
            # each objective.
            if "EVAL" in cap:
                res = 0
                for iN in range(N):
                    res += f.eval(X[:, iN])
                nptest.assert_array_almost_equal(res, f.eval(X))

            # Each column is the prox of one of the N problems.
            if "PROX" in cap:
                res = np.zeros((n, N))
                for iN in range(N):
                    res[:, iN] = f.prox(X[:, iN], step)
                nptest.assert_array_almost_equal(res, f.prox(X, step))

            # Each column is the gradient of one of the N problems.
            if "GRAD" in cap:
                res = np.zeros((n, N))
                for iN in range(N):
                    res[:, iN] = f.grad(X[:, iN])
                nptest.assert_array_almost_equal(res, f.grad(X))

    def test_prox_star(self):
        n = 10
        x = 3 * np.random.randn(n, 1)
        f = functions.norm_l1()
        f2 = functions.dummy()
        f2.prox = lambda x, T: functions._prox_star(f, x, T)
        gamma = np.random.rand()

        p1 = f.prox(x, gamma)
        p2 = functions._prox_star(f2, x, gamma)

        np.testing.assert_array_almost_equal(p1, p2)

        p1 = f.prox(x, gamma) - x
        p2 = -gamma * f2.prox(x / gamma, 1 / gamma)
        np.testing.assert_array_almost_equal(p1, p2)
